package fips.sim.dig.base;

import java.math.BigInteger;
import java.util.ArrayList;

/** The timer synchronizes all simulator components
 */
public class Timer implements Runnable {

/** Listeners for timer events
 */
  ArrayList		listeners;

  Thread		thread;

  BigInteger		time;

  BigInteger		end;

  ArrayList		components;

/** If true the timer can't run faster than realtime.
 *  The start value is used to measure the time really elapsed
 */
  boolean		realtime;


  private final static int TIMER_STARTED = 0;
  private final static int TIMER_STOPPED = 1;
  private final static int TIMER_STEP = 2;



  public Timer() {
    time=BigInteger.ZERO;
    components=new ArrayList();
    listeners=new ArrayList();

    realtime=true;
    thread=null;
  }



// ************************************************************************* //
//
//			Listeners Management				     //
//
// ************************************************************************* //

  public void addTimerListener(TimerListener t) {
    synchronized(listeners) {
      if(t!=null) listeners.add(t);
    }
  }
  public void removeTimerListener(TimerListener t) {
    synchronized(listeners) {
      listeners.remove(t);
    }
  }


  private void fireTimerState(int state) {
    synchronized(listeners) {
      for(int i=0; i<listeners.size(); i++) {
        TimerListener tl=(TimerListener)listeners.get(i);
        switch(state) {
          case TIMER_STARTED:
            tl.timerStarted(time); break;
          case TIMER_STOPPED:
            tl.timerStopped(time); break;
          case TIMER_STEP:
            tl.timerStepComplete(time); break;
          default:
            throw new IllegalArgumentException("State "+state);
        }
      }
    }
  }


// ************************************************************************* //
//
//			Component Management				     //
//
// ************************************************************************* //



  public synchronized void addComponent(fips.sim.dig.base.Component c) {
    if(components.contains(c)) throw new IllegalArgumentException();

/*
    while(requcnt!=components.size()) {
      try {
        this.wait();
      } catch (InterruptedException ie) {}
    }
*/

    components.add(c);
  }


  public synchronized void removeComponent(fips.sim.dig.base.Component c) {
    throw new InternalError();
  }










// ************************************************************************* //
//
//				Thread Part
//
// ************************************************************************* //



  public synchronized void startTimer(BigInteger end) {
    if(thread!=null) {
      throw new IllegalThreadStateException("Timer still running.");
    }

    thread=new Thread(this);
    this.end=end;
    thread.start();
  }

/** Make just one little simulation step. This method
 *  calculates the result in the calling thread so there
 *  is no overhead for creating the asynchronous simulation
 *  thread.
 */
  public synchronized void doStep() {
    //TRACE (".doStep() at "+time);
    if(thread!=null) {
      throw new IllegalThreadStateException("Timer allready running.");
    }
    end=time.add(BigInteger.ONE);
    run();

    //ARM (".doStep() timer now "+time);
  }



  public synchronized boolean isRunning() {
    return(thread!=null);
  }


/** Trigger halt of timer and wait until it is really halted.
 */
  public synchronized void stopTimer() {
    while(thread!=null) {
      end=BigInteger.ZERO;

      try {
        wait();
      } catch (InterruptedException ie) { }
    }
  }


  public synchronized BigInteger getTimerValue() {
    return(time);
  }


  public synchronized BigInteger getTimerEnd() {
    return(end);
  }


  public void run() {
    int				i;
    long			minstep, step;
    BigInteger			ms, bi;

    ms=BigInteger.valueOf(1000000000000L);

    long start=System.currentTimeMillis();
    if(time.bitLength()>0) {
      start-=time.divide(ms).longValue();
    }

    fireTimerState(TIMER_STARTED);

    try {

      while((end==null)||(end.compareTo(time)>0)) {
        //NEVER ("#: starting timer cycle at "+time+"fs");

        minstep=Long.MAX_VALUE;
        for(i=0; i<components.size(); i++) {
          step=((fips.sim.dig.base.Component)components.get(i)).getNextStep();
          if(step<0) {
            throw new InternalError("Step "+step+" of component "+
                 ((fips.sim.dig.base.Component)components.get(i)).getName()+
                 " is invalid.");
          }
          if(step<minstep) minstep=step;
        }

        if(realtime) {
          bi=BigInteger.valueOf(System.currentTimeMillis()-start).
                                multiply(ms).subtract(time);

// bi is now the max step realtime mode allows
          if(bi.signum()<=0) minstep=(minstep==0?0:1);
          else {
            if((bi.bitLength()<64)&&(bi.longValue()<minstep)) {
              minstep=bi.longValue();
              //NEVER ("#: adjusted minstep to "+minstep+
              //NEVER  "fs due to realtime mode.");
            }
          }
        }


        //NEVER ("#: min step is "+minstep+"fs, stepping it.");

        time=time.add(BigInteger.valueOf(minstep));
        for(i=0; i<components.size(); i++) {
          ((fips.sim.dig.base.Component)components.get(i)).doStep(minstep);
        }

        fireTimerState(TIMER_STEP);

        try {
          Thread.sleep(10);
        } catch (InterruptedException ie) {}

      }
    } finally {

      //TRACE ("#: timer main loop finished at "+time);

      synchronized(this) {
        thread=null;

        notifyAll();
      }
      fireTimerState(TIMER_STOPPED);
    }
  }


}

