package fips.sim.dig.comp;

import java.awt.Point;
import java.awt.Dimension;
import java.awt.Graphics;

import java.util.logging.Logger;

public abstract class AbstractComponent implements fips.sim.dig.base.Component {

  String			name;

  private Pin			pins[];
  private int			pinnums[];
  private int			pinoutvals[];


/** Current size of the component, not dependent on the current
 *  orientation or scale of the component.
 */
  protected Dimension		size;

/** Rotated size of the component
 */
  private Dimension		rotsize;

  int				scale;
  Point				offset;
  int				orientation;
  boolean			needsupdate;

  protected Logger		logger;


  public AbstractComponent(int pincnt) {

    logger=Logger.getLogger("digsim.comp");

    name=getComponentName();

    pins=new fips.sim.dig.comp.Pin[pincnt];
    pinnums=new int[pincnt];
    pinoutvals=new int[pincnt];
    for(int i=0; i<pincnt; i++) pinoutvals[i]=Pin.HiZ;
  }


  public abstract String getComponentName();

  public abstract String getDescription();

  public abstract String getPinDescription(int num);


  public String getName() {
    return(name);
  }
 
 
  public void setName(String name) {
    this.name=name;
  }



  public int getPinCount() {
    return(pins.length);
  }


  public void setPin(int num, Pin p) {
    logger.fine(".setPin(num="+num+", pin="+p+") on "+name);

    if(num>=pins.length) throw new IllegalArgumentException();

    if(pins[num]!=null) pins[num].removeComponent(pinnums[num]);
    pins[num]=p;
    if(p!=null) {
      pinnums[num]=p.addComponent();
      p.setValue(pinoutvals[num], pinnums[num]);
    }
  }


  public Pin getPin(int num) {
    if(num>=pins.length) throw new IllegalArgumentException();
    return(pins[num]);
  }




/** Get hints for the max length of the next step
 */
  public abstract long getNextStep();




/** Do the following calculation step. If cycles is too big,
 *  IllegalArgumentException is thrown
 */
  public abstract void doStep(long cycles);




/** Reset cyclecount, params, etc...
 */
  public abstract void reset();



  public abstract void writeState(java.io.ObjectOutputStream oos)
                                  throws java.io.IOException;

  public abstract void readState(java.io.ObjectInputStream ois)
                                 throws java.io.IOException,
                                        ClassNotFoundException;



  protected void setSize(Dimension newsize) {
    size=newsize;
    rotsize=null;
  }

/** Get the GUI size of this component in the current orientation.
 */
  public Dimension getSize() {
    if(rotsize==null) {
      rotsize=new Dimension(size.width, size.height);
      if((orientation&1)!=0) {
        rotsize.width=size.height;
        rotsize.height=size.width;
      }
    }

    return((Dimension)rotsize.clone());
  }


  public void setGUIParameters(Point offset, int scale, int rot) {
    logger.fine(".setGraphics(offset="+offset+
                ", scale="+scale+", rot="+rot+").");

    this.offset=offset;
    this.scale=scale;
    this.orientation=rot;
    rotsize=null;
    needsupdate=true;
  }


  public int getOrientation() {
    return(orientation);
  }


  public boolean needsUpdate() {
    return(needsupdate);
  }

  public abstract void repaint(Graphics g);
  public abstract void update(Graphics g);

  public abstract void doClick(int x, int y, int button);
  public abstract Point getPinPos(int num);


  public Object clone() {
    AbstractComponent ac=null;

    try {
      ac=(AbstractComponent)super.clone();
    } catch (CloneNotSupportedException cnse) {
      cnse.printStackTrace();
      throw new InternalError();
    }

    // AbstractComponent ac=new AbstractComponent(pins.length);

    ac.name=name;
    ac.pins=new Pin[pins.length];
    ac.pinnums=new int[pins.length];
    ac.pinoutvals=(int[])pinoutvals.clone();
    for(int i=0; i<pins.length; i++) {
      if(pins[i]==null) pinnums[i]=-1;
      else {
        ac.pins[i]=pins[i];
        pinnums[i]=pins[i].addComponent();
        ac.pins[i].setValue(pinoutvals[i], i);
      }
    }

    ac.size=size==null?null:(Dimension)size.clone();
    ac.rotsize=rotsize==null?null:(Dimension)rotsize.clone();
    ac.scale=scale;
    ac.offset=offset==null?null:(Point)offset.clone();
    ac.orientation=orientation;
    ac.needsupdate=true;

    return(ac);
  }



/** Calculate the position on the rotated component.
 */
  protected int[] rotatePoint(int x, int y) {
    int ret[]=new int[2];
    switch(orientation) {
      case 0:
        ret[0]=x; ret[1]=y;
        break;
      default:
        throw new InternalError();
    }
    return(ret);
  }




//**********************************
//
//			Extra methods in AbstractComp
//
//*******************************

  protected void setPinCount(int count) {
    if(pins!=null) {
      for(int i=0; i<pins.length; i++) {
        if(pins[i]!=null) {
          throw new InternalError();
        }
      }
    }
 
    pins=new Pin[count];
    pinnums=new int[count];

    int il[]=new int[count];
    System.arraycopy(pinoutvals, 0, il, 0, Math.min(count, pinoutvals.length));
    pinoutvals=il;
  }


  protected int getPinValue(int num) {
    if(pins[num]==null) return(Pin.HiZ);
    else return(pins[num].getValue());
  }


  protected void setPinValue(int val, int num) {
    if(val==pinoutvals[num]) return;

    pinoutvals[num]=val;
    if(pins[num]!=null) pins[num].setValue(val, pinnums[num]);
  }


  public static long getLong(java.math.BigInteger val) {
    if(val.bitLength()<62) return(val.longValue());
    else return(Long.MAX_VALUE);
  }


}



