package fips.sim.dig.comp;

import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Point;
import java.io.IOException;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;

import fips.sim.dig.base.BooleanParameter;
import fips.sim.dig.base.Parameter;
import fips.sim.dig.base.ParameterChangeListener;
import fips.sim.dig.base.SizeParameter;
import fips.sim.dig.base.StringParameter;


/** Component for hardware debugging. Prints message if one
 *  or more pins connected to the debugging component change.
 *  Number of pins, debug message and debug reason parametrizable.
 *
 *  Message structure:
 *  ${time} -> current execution time
 *  ${pin[a]-[b]bdh} -> Print pin data
 *
 *  In clockmode, the component will generate a message
 *  every time the signal reaches HIGH.
 *
 */
public class Debug extends AbstractComponent
                   implements ParameterChangeListener {

  private final static String   paramnames[]={"Size", "Message", "Clockmode"};
  private final static String   paramdescr[]={
          "Size of the debug component",
          "Debug message, ${time [format]} for time, ${pin[a]-[b][format]} for pindata. Valid format strings are b(binary) d(decimal) and h(hex), if none decimal.\nIf the pinvalues should be printed in a fixed-length string, then the number of digits has to follow the format char.\nThe time may contain a char for the lsd (f,p,n,u,m,s) and the number of digits to print.",
          "If true, pin0 is clock and debug message is printed at rising edge"};

  private final static int      PARAM_SIZE  = 0;
  private final static int      PARAM_MESSAGE  = 1;
  private final static int      PARAM_CLOCKMODE  = 2;

  Parameter			params[];


  private final static int	FORMAT_BINARY=0;
  private final static int	FORMAT_DECIMAL=1;
  private final static int	FORMAT_HEX=2;
  private final static int	FORMAT_TIME=3;


  String			debugmsg="Values at ${time n6}: ${pin0,1,3-4d2}";

/** Debugmessage broken into parts. Strings are printed directly,
 *  integer lists contain format in first field, number of
 *  digits in next and list of pin ranges.
 */
  Object			splitmsg[]=null;

  boolean			clockmode=false;

  int				pinvalues[];

  java.math.BigInteger		time;

  private Logger		logger;

  public Debug() {
    super(2);

    logger=Logger.getLogger("digsim.comp");

    time=java.math.BigInteger.ZERO;

    size=new java.awt.Dimension(2, 1);

    pinvalues=new int[2];
  }


  public String getComponentName() {
    return("Debug");
  }

  public String getDescription() {
    return("Component listens for changes on pins and writes customized message if changecondition met.");
  }

  public String getPinDescription(int num) {
    if(num>=getPinCount()) {
      throw new IllegalArgumentException("Only "+getPinCount()+
                                         " pins available.");
    }

    if((clockmode)&&(num==0)) return("Clock pin, action at rising edge");

    return("Debug pin"+num);
  }


  public int getParameterCount() { return(3); }

  public Parameter[] getParameters() {
    if(params==null) {
      params=new Parameter[3];
      for(int i=0; i<3; i++) {
        switch(i) {
          case PARAM_SIZE:
            params[i]=new SizeParameter(paramnames[i], paramdescr[i],
                                        new Dimension(2, 1),
                                        (Dimension)(size.clone()),
                                        new Dimension(2, 1),
                                        new Dimension(100, 100),
                                        true, true, true);
            break;
          case PARAM_MESSAGE:
            params[i]=new StringParameter(paramnames[i], paramdescr[i],
                                          "", debugmsg, Integer.MAX_VALUE);
            break;
          case PARAM_CLOCKMODE:
            params[i]=new BooleanParameter(paramnames[i], paramdescr[i],
                                           false, clockmode);
            break;
          default:
            throw new InternalError();
        }
        params[i].addParameterChangeListener(this);
      }
    }

    return(params);
  }


  public void parameterChanged(Parameter param) {
    int pnum;

    for(pnum=0; pnum<params.length; pnum++) {
      if(params[pnum]==param) break;
    }
    if(pnum==params.length) throw new IllegalArgumentException();

    switch(pnum) {
      case PARAM_SIZE:
// change the size of the component, including pin-map

        logger.severe("not implemented");

        break;
      case PARAM_MESSAGE:
        debugmsg=(String)param.getValue();
        splitmsg=null;
        break;
      case PARAM_CLOCKMODE:
        clockmode=((Boolean)param.getValue()).booleanValue();
        break;
      default:
        throw new InternalError();
    }
  }



  private final static char STATE_CHARS[]={'0', '1', 'u', 'z', '0', '1', 'z'};

  public long getNextStep() {
    boolean debug=false;

    if(clockmode) {
      if((Pin.getReducedState(getPinValue(0))!=Pin.getReducedState(pinvalues[0]))&&
         (Pin.getReducedState(getPinValue(0))==1)) debug=true;
      for(int i=0; i<getPinCount(); i++) pinvalues[i]=getPinValue(i);
    } else {
      for(int i=0; i<getPinCount(); i++) {
        int old=pinvalues[i];
        if(old!=(pinvalues[i]=getPinValue(i))) debug=true;
      }
    }

    if(debug) {
      if(splitmsg==null) splitmsg=splitMessage(debugmsg);

      StringBuffer sbuf=new StringBuffer();
      for(int i=0; i<splitmsg.length; i++) {
        if(splitmsg[i] instanceof String) {
          sbuf.append(splitmsg[i]);
          continue;
        }

        if(splitmsg[i] instanceof int[]) {
          int nums[]=(int[])splitmsg[i];
          String str=null;
          switch(nums[0]) {
            case FORMAT_BINARY:
              byte buf[]=new byte[nums[1]>=0?nums[1]:nums.length-2];
              int j;
              for(j=0; (j<buf.length)&&(j+2<nums.length); j++) {
                buf[buf.length-j-1]=(byte)STATE_CHARS[pinvalues[nums[nums.length-j-1]]];
              }
              if(j!=buf.length) buf[j]=' ';
              str=new String(buf);
              sbuf.append(str);
              break;
            case FORMAT_TIME:
              BigInteger tn=time;
              if(nums[1]!=0) tn=tn.divide(BigInteger.valueOf(10).pow(nums[1]));
              if(nums[2]<0) sbuf.append(tn);
              else {
                str=tn.toString();
                if(str.length()>nums[2]) {
                  str=str.substring(str.length()-nums[2], str.length());
                } else {
                  for(int k=str.length(); k<nums[2]; k++) sbuf.append('0');
                }
                sbuf.append(str);
              }
              switch(nums[1]) {
                case 0: sbuf.append("fs"); break;
                case 3: sbuf.append("ps"); break;
                case 6: sbuf.append("ns"); break;
                case 9: sbuf.append("us"); break;
                case 12: sbuf.append("ms"); break;
                case 15: sbuf.append("s"); break;
                default:
                  sbuf.append("xs");
                  logger.severe("Unsupported format "+nums[1]);
                  break;
              }
              break;
            default:
              throw new IllegalArgumentException("Unsupported output format "+nums[0]);
          }

          continue;
        }

        throw new InternalError("Unknown val "+splitmsg[i]);
      }

      System.out.println(sbuf.toString());
    }
    return(Long.MAX_VALUE);
  }



/** Do the following calculation step.
 *  @param cycles do calculation steps.
 *  @throws IllegalArgumentException if cycles is to big
 */
  public void doStep(long cycles) {
    time=time.add(java.math.BigInteger.valueOf(cycles));
  }




/** Reset cyclecount, params, etc...
 */
  public void reset() {
    time=java.math.BigInteger.ZERO;
    throw new InternalError();
  }




  public void writeState(java.io.ObjectOutputStream oos)
                         throws IOException {
    oos.writeInt(getPinCount());
    oos.writeObject(debugmsg);
    oos.writeBoolean(clockmode);
    oos.writeObject(pinvalues);
    oos.writeObject(time);
  }
 
 
  public void readState(java.io.ObjectInputStream ois)
                        throws IOException {
    try {
      setPinCount(ois.readInt());
      debugmsg=(String)ois.readObject();
      splitmsg=null;
      clockmode=ois.readBoolean();
      pinvalues=(int[])ois.readObject();
      time=(BigInteger)ois.readObject();
    } catch (ClassNotFoundException cnfe) {
      cnfe.printStackTrace();
      throw new IOException("Can't read state: "+cnfe);
    }
  }




  public void repaint(Graphics g) {
    g.setColor(java.awt.Color.white);
    java.awt.Dimension rotsize=getSize();
    g.drawRect(offset.x+1, offset.y+1, (rotsize.width<<scale)-1,
                                          (rotsize.height<<scale)-1);
    needsupdate=false;
  }

  public void update(Graphics g) {
    needsupdate=false;
  }


  public void doClick(int x, int y, int button) {
  }

/** Get the pin position of a component. The position is
 *  relative to the origin of the component and is affected
 *  by the rotation parameter.
 */
  public Point getPinPos(int num) {
    logger.finer(".getPinPos(num="+num+").");

    int pins=getPinCount();
    if(num>=pins) throw new IllegalArgumentException();

    Point p=new Point();
    int offs=0;
    if(num<(pins>>1)) {
      p.x=num+1; p.y=0;
    } else {
      p.x=num-(pins>>1)+1; p.y=size.height;
    }

    if((orientation&1)!=0) {
      throw new InternalError();
    }

    logger.finest("Pos is "+p.x+"x"+p.y);

    return(p);
  }




  public Object clone() {
    Debug d = (Debug)super.clone();
    d.debugmsg=debugmsg;
    d.clockmode=clockmode;
    return(d);
  }



/** Break up the debug message string into its parts and
 *  store pinnumbers as integer arrays.
 */
  private Object[] splitMessage(String msg) {
    ArrayList list=new ArrayList();

    for(int pos=0; pos<msg.length(); ) {
      int end=msg.indexOf("${", pos);
      if(end<0) {
        list.add(msg.substring(pos));
        break;
      }
      if(end!=pos) {
        list.add(msg.substring(pos, end));
        pos=end;
        continue;
      }

      int close=msg.indexOf('}', end+2);
      if(close<0) {
        list.add(msg.substring(pos));
        break;
      }

      int nextopen=msg.indexOf("${", end+2);
      if((nextopen>0)&&(nextopen<close)) {
        list.add(msg.substring(pos, nextopen));
        pos=nextopen;
        continue;
      }

      String str=msg.substring(pos+2, close);
      //ARM ("#: analyzing \""+str+"\".");

      if(str.startsWith("time")) {
        int il[]=new int[3];
        il[0]=FORMAT_TIME;
        if(str.trim().equals("time")) {
          il[1]=0;	// start at fs
          il[2]=-1;	// print all digits
        } else {
          str=str.substring(4).trim();

          boolean foundsc=false;
          char c=str.charAt(0);
          if((!Character.isDigit(c))&&(c!='-')&&(c!='+')) {
            str=str.substring(1);
            foundsc=true;
          } else {
            c=str.charAt(str.length()-1);
            if((!Character.isDigit(c))&&(c!='-')&&(c!='+')) {
              str=str.substring(0, str.length()-1);
              foundsc=true;
            }
          }

          if(foundsc) {
            switch(c) {
              case 'f': il[1]=0; break;
              case 'p': il[1]=3; break;
              case 'n': il[1]=6; break;
              case 'u': il[1]=9; break;
              case 'm': il[1]=12; break;
              case 's': il[1]=15; break;
              default:
                il=null;
            }
          }

          if((il!=null)&&(str.trim().length()>0)) {
            try {
              il[2]=Integer.parseInt(str.trim());
            } catch (NumberFormatException nfe) {
              il=null;
            }
          }
        }

        if(il==null) list.add("syntax error");
        else list.add(il);

      } else if(str.startsWith("pin")) {
        int pinlist[]=new int[256];
        pinlist[0]=FORMAT_BINARY;
        pinlist[1]=-1;
        int pincnt=2;

        int spos;
        for(spos=3; spos<str.length(); ) {
          char c=str.charAt(spos);
          if((c>='0')&&(c<='9')) {
            int endp=spos+1;
            while((endp<str.length())&&(Character.isDigit(str.charAt(endp)))) endp++;

            try {
              pinlist[pincnt++]=Integer.parseInt(str.substring(spos, endp));
            } catch (NumberFormatException nfe) {
              break;
            }

            spos=endp;
            if(spos==str.length()) {
              break;
            }

            c=str.charAt(spos);
            if(c=='-') {
//range
              //ERROR ("Range not yet");
              break;
            } else if(c==',') {
              if((spos+1==str.length())||
                 (!Character.isDigit(str.charAt(spos+1)))) break;
              spos++;
            } else {
// unknown char, perhaps format char
              continue;
            }
          } else {
// must be format char, only followed by digit num
            if(c=='d') pinlist[0]=FORMAT_DECIMAL;
            else if(c=='b') pinlist[0]=FORMAT_BINARY;
            else if(c=='h') pinlist[0]=FORMAT_HEX;
            else break;
            spos++;

            if(spos!=str.length()) {
              try {
                pinlist[1]=Integer.parseInt(str.substring(spos));
              } catch (NumberFormatException nfe) {
                break;
              }
            }
          }
        }

        if(spos<str.length()) {
// parsing failed, just add error string
          list.add("syntax error");
        } else {
          int il[]=new int[pincnt];
          System.arraycopy(pinlist, 0, il, 0, pincnt);
          list.add(il);
        }
      } else {
        //TRACE ("#: unparseable string "+str);
        list.add(msg.substring(pos, close+1));
      }

      pos=close+1;
    }

    return(list.toArray());
  }



}




