package fips.sim.dig.comp;

import java.awt.Graphics;
import java.util.logging.Level;
import java.util.logging.Logger;

import fips.sim.dig.base.Parameter;
import fips.sim.dig.base.ByteArrayParameter;



/** Simulate Gal16v8 chip including unknown reset state
 *  and signal propagation through fuse array.
 *
 *  pinvalues, regvalues are reduced pin values, they
 *  may be 0, 1 or -1
 */
public class Gal16V8 extends AbstractComponent {


  private final static String	paramnames[]={"Fusedata"};
  private final static String	paramdescr[]={"Internal fuse map of the gal component\nThe bit 0 is the lsb of the first byte so 3FC2 is 11111100 01000011\n Fuses 0-2047 (byte 0..255) and array\n     2048-2055 (byte 256) xorfuses\n    2056-2119 (byte 101h..108h) signature\n    2120-2127 (byte 109h) ac0 fuses\n     2128-2191 (byte 10Ah..111h) ptd fuses\n    2192 (112.0h)  syn fuse\n   2193 (112.1h) ac0 fuse\n\nGet a gal16v8 datasheet to see what these fuses are for.\n"};


  private final static int	PARAM_FUSEDATA	= 0;


/** Pin positions for the virtual gal16v8 chip, not
 *  identic with real pinout!!
 */
  private final static int	CLK_PIN		=  0;
  private final static int	IN0_PIN		=  1;
  private final static int	IO0_PIN		=  9; // is pin 19 on DIL!!
  private final static int	OE_PIN		= 17;

  private final static int	SIMPLEMODE_INPUTMAP[]={
	1,0, 2,9, 3,10, 4,11, 5,14, 6,15, 7,16, 8,17 };


  private final static int	FIRST_AND_FUSE		=    0;
  private final static int	LAST_AND_FUSE		= 2047;


  private final static int	XOR0_FUSE		= 2048;
  private final static int	XOR7_FUSE		= 2055;

  private final static int	SIGNATURE0_FUSE		= 2056;
  private final static int	SIGNATURE63_FUSE	= 2119;

  private final static int	AC1_0_FUSE		= 2120;
  private final static int	AC1_7_FUSE		= 2127;

// positions for product term disable fuses
  private final static int	PTD0_FUSE		= 2128;
  private final static int	PTD63_FUSE		= 2191;

  private final static int	SYN_FUSE		= 2192;
  private final static int	AC0_FUSE		= 2193;

  private final static int	FUSE_BYTES		= 275;


// security cell???

// syn=0 ac0=1 ac1=0 -> reg-conf reg-mode		// sg0==syn
// syn=0 ac0=1 ac1=1 -> comb-conf reg-mode
// syn=1 ac0=1 ac1=1 -> combIO-conf complex-mode	// pins 13-18
// syn=1 ac0=1 ac1=1 -> combIO-conf complex-mode	// pins 12, 19
// syn=1 ac0=0 ac1=0 -> combO-conf simple-mode
// syn=1 ac0=0 ac1=1 -> input-conf simple-mode


/*
Feedback syn ac0 ac1
010	/Q
011	Pin
111	Pin 13-18
111	Pin 12,19
100	Pin 12-14 17-19
100	Pin 15,16
101	Pin 12-14 17-19

	10	11	01
Simple Mode	Registered mode
	Complex Mode
FB0	1	1	19
FB1	19	18	18
FB2	18	17	17
FB3	17	16	16
FB4	14	15	15
FB5	13	14	14
FB6	12	13	13
FB7	11	11	12
*/


// timing default for Gal16V8-15ns

  private long			COMBINATORIAL_PROPAG	= 15000000;
  private long			REGISTERED_MINSETUP	= 12000000;
  private long			REGISTERED_MINHOLDTIME	=        0;
  private long			REGISTERED_MINCLOCK	= 10000000;
  private long			MINDCLOCK		= 24000000;
  private long			REGISTERED_CLOCK2OUT	= 12000000;
  private long			REGISTERED_FEEDBACK	=  8000000;
  private long			CLOCK2COMBINATOUT	= 25000000;
  private long			OECHANGE		= 15000000;


/** The value this component wants to drive to the pin.
 *  The array has the length 18 although only 8 pins are
 *  able to drive the output pins. The pin values are full
 *  states and HiZ for input pins.
 */
  int				pinoutvalues[];

/** Reduced values at pin inputs. The reduction is done when
 *  the pins are read in. -1 denotes any unknown state, 0 and
 *  1 the known ones.
 */
  int				pinvalues[];

/** Time since last change of pin. Every step that pin
 *  stays the same, the step width is added to the time
 *  until Long.MAX_VALUE reached.
 */
  long				lastinputchanges[];


/** Reduced values in the register.
 */
  int				regvalues[];

/** Time of the last clockchange. If change at reg[x] happened
 *  more than REGISTERED_CLOCK2OUT fs ago, output values get
 *  correct value.
 *  Feedback lines will receive value after REGISTERED_FEEDBACK fs.
 */
  long				lastregchanges[];


/** Reduced state of clock before change.*/
  int				lastclkval;

/** Time when lastclkchange left state lastclkval.*/
  long				lastclkchange;



/** Time of next change. A value of >= 0 indicates that the
 *  component will change outputs without any input change.
 */
  long				nextchange;

/** Values to be written to output after nextchange time,
 *  lenght 18. The values are normal pin states not reduced
 *  states!.
 */
  int				nextvalues[];

/** Next register content, values are reduced states.
 */
  int				nextregvalues[];


  byte				fusedata[];

  java.awt.Color		border=new java.awt.Color(0xC0C0C0);


  Parameter			params[];

/** Names of the pins
 */ 
  String			pinnames[];

  public Gal16V8() {
    super(18);

    logger=Logger.getLogger("digsim.comp.Gal16V8");

    size=new java.awt.Dimension(6, 20);

    fusedata=new byte[FUSE_BYTES];

    pinvalues=new int[18];
    lastinputchanges=new long[18];
    pinoutvalues=new int[18];
    nextvalues=new int[18];

    for(int i=0; i<18; i++) {
      pinoutvalues[i]=Pin.HiZ;
      setPinValue(pinoutvalues[i], i);
      pinvalues[i]=-1;
    }

    regvalues=new int[8];
    nextregvalues=new int[8];
    lastregchanges=new long[8];

    for(int i=0; i<8; i++) {
      regvalues[i]=Pin.UNKNOWN;
      nextregvalues[i]=-1;
    }



// trigger recalculation of output pins and write them to pins
    doArray();
  }


  public String getComponentName() {
    return("Gal16V8");
  }

  public String getDescription() {
    return("Standard gate array logic with 8 macrocells.");
  }


  public String getPinDescription(int num) {
    throw new InternalError();
  }


  public int getParameterCount() { return(paramnames.length); }

  public fips.sim.dig.base.Parameter[] getParameters() {
    if(params==null) {
      params=new Parameter[paramnames.length];
      for(int i=0; i<paramnames.length; i++) {
        switch(i) {
          case PARAM_FUSEDATA:
            params[i]=new ByteArrayParameter(paramnames[i], paramdescr[i],
                                             new byte[FUSE_BYTES], fusedata,
                                             FUSE_BYTES, FUSE_BYTES);
            break;
          default:
            throw new InternalError("Inval param "+i);
        }
      }
    }

    return(params);
  }


  public void setParameter(int num, Object val) {
    //TRACE (".setParameter(num="+num+", val="+val+").");

    switch(num) {
      case PARAM_FUSEDATA:
        byte buf[]=(byte[])val;
        if(buf.length!=FUSE_BYTES) throw new IllegalArgumentException();
        fusedata=buf;
        params[num]=new ByteArrayParameter(paramnames[num], paramdescr[num],
                                           new byte[FUSE_BYTES], fusedata,
                                           FUSE_BYTES, FUSE_BYTES);
        break;
      default:
        throw new InternalError();
    }
  }





  public void setPin(int num, Pin p) {
    super.setPin(num, p);

    setPinValue(pinoutvalues[num], num);
    int i=Pin.getReducedState(getPinValue(num));

    if(pinvalues[num]!=i) {
// now connected to other potential, treat like
// potential change since last doStep
      lastinputchanges[num]=0;

    }
  }





/** Get hints for the max length of the next step
 */
  public long getNextStep() {
    //NEVER (".getNextStep()");

    boolean changed=true;
    for(int i=0; i<18; i++) {
      int val=Pin.getReducedState(getPinValue(i));

      if(val!=pinvalues[i]) {
        changed=true;
        pinvalues[i]=val;
      }
    }

    if(changed) {
      doArray();
    }

    return(nextchange);
  }




/** Do the following calculation step. If cycles is too big,
 *  IllegalArgumentException is thrown
 */
  public void doStep(long cycles) {
    logger.finer(".doStep("+cycles+").");

    nextchange-=cycles;
    if(nextchange<0) throw new InternalError(); // just for sure -> testing

    if(nextchange==0) {
      for(int i=0; i<18; i++) {
        if(nextvalues[i]==pinoutvalues[i]) continue;
        setPinValue(nextvalues[i], i);
        pinoutvalues[i]=nextvalues[i];
      }
      nextchange=-1;
    }

    for(int i=0; i<18; i++) {
      lastinputchanges[i]+=cycles;
      if(lastinputchanges[i]<0) lastinputchanges[i]=Long.MAX_VALUE;
    }

    for(int i=0; i<8; i++) {
      lastregchanges[i]+=cycles;
      if(lastregchanges[i]<0) lastregchanges[i]=Long.MAX_VALUE;
    }
  }




/** Reset cyclecount, params, etc...
 */
  public void reset() {
    throw new InternalError();
  }



/** Serialize all statevars.
 */
  public void writeState(java.io.ObjectOutputStream oos)
                         throws java.io.IOException {
    oos.writeObject(pinoutvalues);
    oos.writeObject(pinvalues);
    oos.writeObject(lastinputchanges);
    oos.writeObject(regvalues);
    oos.writeObject(lastregchanges);
    oos.writeLong(nextchange);
    oos.writeObject(nextvalues);
    oos.writeObject(nextregvalues);
    oos.writeObject(fusedata);
  }



  public void readState(java.io.ObjectInputStream ois) 
                        throws java.io.IOException, ClassNotFoundException {
    pinoutvalues=(int[])ois.readObject();
    pinvalues=(int[])ois.readObject();
    lastinputchanges=(long[])ois.readObject();
    regvalues=(int[])ois.readObject();
    lastregchanges=(long[])ois.readObject();
    nextchange=ois.readLong();
    nextvalues=(int[])ois.readObject();
    nextregvalues=(int[])ois.readObject();
    fusedata=(byte[])ois.readObject();
  }




/* Paint the component with small
 */
  public void repaint(Graphics graphics) {
    java.awt.Dimension size=getSize();
    graphics.setColor(java.awt.Color.black);
    graphics.fillRect(offset.x, offset.y,
                      size.width<<scale, size.height<<scale);

    graphics.setColor(java.awt.Color.white);
    if((orientation&1)==0) {
      if(scale<3) {
        graphics.drawRect(offset.x+1, offset.y+1,
                          (size.width<<scale)-2, (size.height<<scale)-2);
      } else {
        graphics.drawRect(offset.x+(1<<(scale-1)), offset.y,
                          (size.width-1)<<scale, size.height<<scale);
        for(int i=0; i<10; i++) {
          graphics.drawRect(offset.x, offset.y+((i*16+5)<<(scale-3)),
                            1<<(scale-1), 3<<(scale-2));

          graphics.drawRect(offset.x+((size.width*2-1)<<(scale-1)),
                            offset.y+((i*16+5)<<(scale-3)),
                            1<<(scale-1), 3<<(scale-2));
        }
      }
    } else {
      throw new InternalError();
    }
    needsupdate=false;
  }


/** Do nothing, component display won't change with state.
 */
  public void update(Graphics g) {
    needsupdate=false;
  }



  public void doClick(int x, int y, int button) {
    //ERROR (".doClick() not yet implemented");
  }

  public java.awt.Point getPinPos(int num) {
    if((num<0)||(num>=18)) throw new IllegalArgumentException();

    java.awt.Point ret=new java.awt.Point();

    switch(orientation) {
      case 0:
        if(num<9) { ret.x=0; ret.y=(num<<1)+1; }
        else { ret.x=6; ret.y=((num-9)<<1)+3; }
        break;
        
      default:
        throw new InternalError();
    }
    return(ret);
  }



  public Object clone() {
    Gal16V8 clone = new Gal16V8();

// now copy all internal vars
    clone.fusedata=(byte[])fusedata.clone();

    return(clone);
  }



// ************************************************************************* //
//
//			Internal Methods				     //
//
// ************************************************************************* //



/** Calculate line values in the current array.
 *  The method stores the next register values and next
 *  pinout values in arrays.
 *  @returns the time until change will occur. If time is 0
 *  the values from nextvalues may be copied to output pins
 *  at doStep
 */
  private long doArray() {
    logger.fine(".doArray()");

    boolean syn=getBit(SYN_FUSE);
    boolean ac0=getBit(AC0_FUSE);

    nextchange=Long.MAX_VALUE;

// get the pots of the lines in the array
// the lines have reduces states 0,1,-1
    int arraylines[]=new int[32];

    if((!syn)&&(!ac0)) {
// unimplemented mode, arraylines may have any state, do nothing

    } else if(syn&&(!ac0)) {	// simple mode
      logger.finer("Calculating in simple_mode");

      for(int i=0; i<16; i++) {
        int pos=SIMPLEMODE_INPUTMAP[i];

        if(lastinputchanges[pos]<COMBINATORIAL_PROPAG) {
          arraylines[i<<1]=-1;
          if(COMBINATORIAL_PROPAG-lastinputchanges[pos]<nextchange) {
            nextchange=COMBINATORIAL_PROPAG-lastinputchanges[pos];
          }
        } else arraylines[i<<1]=pinvalues[pos];


      }
    } else if((!syn)&&ac0) {
      logger.finer("Running in registered mode");

// configuration for registered mode

// fill the input-only arraylines 
      for(int i=0; i<8; i++) {
        if(lastinputchanges[i+1]<COMBINATORIAL_PROPAG) {
          arraylines[i<<2]=-1;
          if(COMBINATORIAL_PROPAG-lastinputchanges[i+1]<nextchange) {
            nextchange=COMBINATORIAL_PROPAG-lastinputchanges[i+1];
          }
        } else arraylines[i<<2]=pinvalues[i+1];
      }

// fill the io/pin lines
      for(int i=0; i<8; i++) {
        if(getBit(AC1_0_FUSE+i)) { // take the pin value
          if(lastinputchanges[i+9]<COMBINATORIAL_PROPAG) {
            arraylines[(i<<2)+2]=-1;
            if(COMBINATORIAL_PROPAG-lastinputchanges[i+9]<nextchange) {
              nextchange=COMBINATORIAL_PROPAG-lastinputchanges[i+9];
            }
          } else arraylines[(i<<2)+2]=pinvalues[i+9];
        } else {

// use content from register, but use inverted output

          if(lastregchanges[i]<REGISTERED_FEEDBACK) {
            arraylines[(i<<2)+2]=-1;
            if(REGISTERED_FEEDBACK-lastregchanges[i]<nextchange) {
              nextchange=REGISTERED_FEEDBACK-lastregchanges[i];
            }
          } else {
            arraylines[(i<<2)+2]=(regvalues[i]==-1)?-1:regvalues[i]^1;
          }
        }
      }


    } else {
      throw new InternalError("Mode with syn="+syn+", ac0="+ac0+" not impl.");
    }


// complement the arraylines
    for(int i=0; i<16; i++) {
      arraylines[(i<<1)+1]=(arraylines[i<<1]==-1)?-1:arraylines[i<<1]^0x1;
    }

    logger.finer("Got arraylines, reduced nextchange to "+nextchange);
/*
    if(logger.wouldLog(Level.FINEST)) {
      StringBuffer sbuf=new StringBuffer();
      sbuf.append("Arraylines: ");
      for(int i=0; i<arraylines.length; i++) {
        sbuf.append(arraylines[i]).append(", ");
      }
      logger.log(sbuf);
    }
*/

// Calculate the 64 products. The product lines have only
// reduced states 0, 1, -1
    int products[]=new int[64];
    for(int i=0; i<64; i++) {
      if(!getBit(PTD0_FUSE+i)) {
        products[i]=0;
        logger.finest("Term "+i+" is disabled");
        continue; // product term disabled, leave 0
      }

      products[i]=1;
      for(int j=0; j<32; j++) {
        if(!getBit(FIRST_AND_FUSE+(i<<5)+j)) {
          products[i]=arraylines[j];
          if(products[i]==0) break;	// can't get 1 any more
        }
      }
      logger.finest("Product "+i+" is "+products[i]);
    }

    for(int i=0; i<8; i++) nextregvalues[i]=-1;

// set nextvalues to -1 to yield exception if not setup properly
    for(int i=0; i<18; i++) nextvalues[i]=-1;

// now calculate the outputs
    if((!syn)&&(!ac0)) {
// illegal hw config, make outs tristate??
      for(int i=0; i<18; i++) nextvalues[i]=Pin.HiZ;

    } else if(syn&&(!ac0)) {
// simple mode, all inputs tristate, no tristate output pins
      for(int i=0; i<IO0_PIN; i++) nextvalues[i]=Pin.HiZ;
      nextvalues[OE_PIN]=Pin.HiZ;

      for(int i=0; i<8; i++) {

        if((i!=3)&&(i!=4)&&(getBit(AC1_0_FUSE+i))) {
// dedicated input mode, not for pin3,4
          nextvalues[IO0_PIN+i]=Pin.HiZ;
        } else {
          nextvalues[IO0_PIN+i]=0;
          for(int j=0; j<8; j++) {
            if(products[(i<<3)+j]!=0) {
              nextvalues[IO0_PIN+i]=products[(i<<3)+j];
              if(nextvalues[IO0_PIN+i]==1) break;
            }
          }
// now nextvalues has reduced pinstate 
// apply the xor-fuse and make full pinstate
          if(!getBit(XOR0_FUSE+i)) {
            nextvalues[IO0_PIN+i]=(nextvalues[IO0_PIN+i]==-1?-1:nextvalues[IO0_PIN+i]^1);
          }
          nextvalues[IO0_PIN+i]=(nextvalues[IO0_PIN+i]==-1?Pin.UNKNOWN:nextvalues[IO0_PIN+i]);
        }

      }
    } else if((!syn)&&ac0) {
// registered mode

      for(int i=0; i<IO0_PIN; i++) nextvalues[i]=Pin.HiZ;
      nextvalues[OE_PIN]=Pin.HiZ;

// if doclock=1 perform regular clock cycle, if -1 malformed clockcycle
// after to slow or short clock pulse
      int doclock=0;

// if 0 don't output, if 1 output, if -1 unknown
      int oeenb=-1;
      if(lastinputchanges[17]>=OECHANGE) {
        oeenb=(pinvalues[17]==-1)?-1:pinvalues[17]^1;
      } else {
        nextchange=Math.min(nextchange, OECHANGE-lastinputchanges[17]); 
      }


      for(int i=0; i<8; i++) {
        int val=0;

// If not in inputmode, or-gate has 8 inputs
        if(!getBit(AC1_0_FUSE+i)) val=products[(i<<3)];

        for(int prodcnt=1; prodcnt<8; prodcnt++) {
          if(val==1) break;
          if(products[(i<<3)+prodcnt]!=0) val=products[(i<<3)+prodcnt];
        }

        if(!getBit(XOR0_FUSE+i)) val=(val==-1?-1:val^1);

        if(getBit(AC1_0_FUSE+i)) {
// inputmode, register value unknown, output depending on product 0

          if(doclock!=0) nextregvalues[i]=-1;
          switch(products[i<<3]) {
            case 0: nextvalues[IO0_PIN+i]=Pin.HiZ; break;
            case 1: nextvalues[IO0_PIN+i]=(val<0?Pin.UNKNOWN:val); break;
            case -1: nextvalues[IO0_PIN+i]=Pin.UNKNOWN; break;
          }
        } else {
// real registered mode
          switch(doclock) {
            case 1: nextregvalues[i]=val; break;
            case -1:
              if(regvalues[i]!=val) nextregvalues[i]=-1;
              break;
          }

          switch(oeenb) {
            case 0: nextvalues[IO0_PIN+i]=Pin.HiZ; break;
            case 1:
              if(lastregchanges[i]<REGISTERED_CLOCK2OUT) {
                nextvalues[IO0_PIN+i]=Pin.UNKNOWN;
                nextchange=Math.min(nextchange,
                                    REGISTERED_CLOCK2OUT-lastregchanges[i]);
              } else nextvalues[IO0_PIN+i]=(regvalues[i]==-1)?Pin.UNKNOWN:regvalues[i]^1; 
              break;
            case -1: nextvalues[IO0_PIN+i]=Pin.UNKNOWN; break;
          }
        }
      }
    } else {
      throw new InternalError("Mode with syn="+syn+", ac0="+ac0+" not impl.");
    }





// compare the nextregvalues with current values
// if different set nextchange to 0

    for(int i=0; i<8; i++) {
      if((nextregvalues[i]!=-1)&&(nextregvalues[i]!=regvalues[i])) {
        nextchange=0;
        break;
      }
    }
    for(int i=0; i<18; i++) {
      if(nextvalues[i]!=pinoutvalues[i]) {
        nextchange=0;
        break;
      }
    }

    logger.finer("Nextchange will be "+nextchange);
    return(nextchange);
  }



  private boolean getBit(int pos) {
    return((fusedata[pos>>3]&(1<<(pos&0x7)))!=0);
  }

}



