package fips.sim.dig.comp;


import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Point;
import java.math.BigInteger;

import java.util.logging.Logger;

import fips.sim.dig.base.Parameter;
import fips.sim.dig.base.ParameterChangeListener;
import fips.sim.dig.base.TimeParameter;
import fips.sim.dig.util.GraphicsUtil;



/** Inverter component
 *  If the input level was stable for certain time, output
 *  will be the inverted value of input.
 *  
 */
public class Inverter extends AbstractComponent
                      implements ParameterChangeListener{

  private final static String	paramnames[]={"Propagation"};
  private final static String	paramdescr[]={"Time from input change to stable output\n"};
 
  private final static int	PARAM_PROPAGATION  = 0;
  private final static BigInteger	default_propag=TimeParameter.TIME_NS;

  private Parameter		params[];


  BigInteger		time;
  BigInteger		stablewait;
  BigInteger		nextstable;
  int			laststate;	// reduced last state

  private Logger	logger;

/** Zero argument constructor for the inverter, needed by editor.
 */
  public Inverter() {
    super(2);

    logger=Logger.getLogger("digsim.comp");

    time=BigInteger.ZERO;
    stablewait=default_propag;
    nextstable=stablewait;
    laststate=-2;

    size=new Dimension(2, 2);
  }


/** Get the short name of the component.
 */
  public String getComponentName() {
    return("Inverter");
  }

  public String getDescription() {
    return("Inverts a signal with configurable propagation.");
  }



  public String getPinDescription(int num) {
    switch(num) {
      case 0: return("Input pin");
      case 1: return("Output pin");
      default:
        throw new IllegalArgumentException("No description for pin "+num+
                   ", only "+getPinCount()+ " pins available.");
    }
  }


  public int getParameterCount() { return(paramnames.length); }


/** Get the parameter descriptions, inverter has only 
 *  propagation time parameter.
 */
  public Parameter[] getParameters() {
    if(params==null) {
      params=new Parameter[paramnames.length];
      for(int i=0; i<paramnames.length; i++) {
        switch(i) {
          case PARAM_PROPAGATION:
            params[i]=new TimeParameter(paramnames[i], paramdescr[i],
                                        default_propag, stablewait,
                                        BigInteger.ZERO, null); 
            break;
          default:
            throw new InternalError("Inval param "+i);
        }

        params[i].addParameterChangeListener(this);
      }
    }
    return(params);
  }


/** Set the parameter.
 */
  public void parameterChanged(Parameter param) {
    int pnum;
    for(pnum=0; pnum<params.length; pnum++) {
      if(param==params[pnum]) break;
    }
    if(pnum==params.length) throw new IllegalArgumentException();

    switch(pnum) { 
      case PARAM_PROPAGATION:
        stablewait=(BigInteger)param.getValue();
        break;
      default:
        throw new InternalError();
    }
  }




/** Get the time when inverter will change its output
 *  pin.
 */
  public long getNextStep() {
    if(laststate!=Pin.getReducedState(getPinValue(0))) {
      nextstable=time.add(stablewait);
      laststate=Pin.getReducedState(getPinValue(0));
      return(0);
    }

    if(laststate==-1) return(Long.MAX_VALUE);

    if(nextstable.compareTo(time)<=0) return(Long.MAX_VALUE);
    return(getLong(nextstable.subtract(time)));
  }



/** Step forward.
 *  @param cycles do calculation steps.
 *  @throws IllegalArgumentException if cycles is to big
 */
  public void doStep(long cycles) {
    time=time.add(BigInteger.valueOf(cycles));

    int i=time.compareTo(nextstable);
    if(i>0) return;

    if((i<0)||(laststate==-1)) setPinValue(Pin.UNKNOWN, 1);
    else {
      logger.finer("Setting output to "+(laststate^1));
      setPinValue(laststate^1, 1);
    }
  }




/** Reset cyclecount, params, etc...
 */
  public void reset() {
    time=java.math.BigInteger.ZERO;
    laststate=-2;
  }



/** Write the inverter's state to the output stream.
 */
  public void writeState(java.io.ObjectOutputStream oos)
                         throws java.io.IOException {
    oos.writeObject(time);
    oos.writeObject(stablewait);
    oos.writeObject(nextstable);
    oos.writeInt(laststate);
  }
 

/** Reconstruct the inverter's state from inputstream
 */
  public void readState(java.io.ObjectInputStream ois)
                        throws java.io.IOException {
    try {
      time=(BigInteger)ois.readObject();
      stablewait=(BigInteger)ois.readObject();
      nextstable=(BigInteger)ois.readObject();
      int laststate=ois.readInt();
      if((laststate<0)||(time.compareTo(nextstable)<0)) setPinValue(Pin.UNKNOWN, 1);
      else setPinValue(laststate, 1);
    } catch (ClassNotFoundException cnfe) {
      cnfe.printStackTrace();
      throw new java.io.IOException(cnfe.toString());
    }
  }




  public void repaint(Graphics graphics) {
    graphics.setColor(Color.black);
    Dimension rotsize=getSize();
    rotsize.width<<=scale;
    rotsize.height<<=scale;
   
    graphics.fillRect(offset.x+1, offset.y+1,
                      rotsize.width-1, rotsize.height-1);

    graphics.setColor(Color.white);
    if(scale<3) {
      graphics.drawArc(offset.x, offset.y, rotsize.width, rotsize.height,
                       0, 360);
    } else {
      GraphicsUtil.drawLine(graphics, orientation, offset, rotsize, 0, 1<<scale,
                            1<<(scale-1), 1<<scale);
      GraphicsUtil.drawLine(graphics, orientation, offset, rotsize,
                            1<<(scale-1), 1<<(scale-1),
                            1<<(scale-1), 3<<(scale-1));
      GraphicsUtil.drawLine(graphics, orientation, offset, rotsize,
                            1<<(scale-1), 1<<(scale-1),
                            3<<(scale-1), 1<<scale);
      GraphicsUtil.drawLine(graphics, orientation, offset, rotsize,
                            1<<(scale-1), 3<<(scale-1),
                            3<<(scale-1), 1<<scale);

      int arcsize=(1<<scale)/3;
      GraphicsUtil.drawArc(graphics, orientation, offset, rotsize,
                           3<<(scale-1), (1<<scale)-arcsize/2,
                           arcsize, arcsize, 0,360);
      GraphicsUtil.drawLine(graphics, orientation, offset, rotsize,
                            (3<<(scale-1))+arcsize, 1<<scale,
                            2<<scale, 1<<scale);
    }
    needsupdate=false;
  }


/** Update the component if changed.
 *  The inverter image won't change so no impl needed
 */
  public void update(Graphics g) {
    needsupdate=false;
  }




/** Method called by editor if component was clicked.
 *  The inverter is not sesitive to clicks, so nothing is
 *  done.
 */
  public void doClick(int x, int y, int button) {
    logger.fine(".doClick() on "+getName());
  }


/** Let the editor know where the connection terminals
 *  are.
 */
  public Point getPinPos(int num) {
    logger.fine(".getPinPos(num="+num+").");

    if(num>=getPinCount()) throw new IllegalArgumentException();

    Point p=new Point();
    p.x=num<<1; p.y=1;

    if(orientation!=0) throw new InternalError();

    logger.finer("Pos of pin "+num+" is "+p.x+"x"+p.y);
    return(p);
  }




  public Object clone() {
    Inverter i = (Inverter)super.clone();

    i.time=time;
    i.stablewait=stablewait;
    i.nextstable=nextstable;
    i.laststate=laststate;
    return(i);
  }



}



