package fips.sim.dig.comp;

import fips.sim.dig.comp.fsm.*;


/**Finite state machine part of the lcd controller. This class
 * just keeps track on the incomming signals and creates error
 * messages if the signal specifications get violated. <Br>
 * Address lines and command are sampled when clock rises,
 * changes during execution of a command results in errors but
 * doesn't bring the controller into an unknown state as the
 * real circuit would.
 */
public class LCDControler extends FSMachine {
  private LCDDisplayGUI		display;



  final static int			VARS_CONTRAST	=  0;
  final static int			VARS_HILIGHT	=  1;
  final static int			VARS_RW		=  2;
  final static int			TIMERS_RW	=  2;
  final static int			VARS_RS		=  5;
  final static int			TIMERS_RS	=  3;


/** Maximum length of an unknown signal on a pin until a state
    may get triggered. The default value is 1 ns. 
 */
  private long			spike=1000000;



/** Time for a pinchange on the contrast pin to show effects.
    The default value is 100ns;
 */
  long					contrwait=100000000L;


/** Time for a pinchange on the hilight pin to show effects.
 *  The default value is 100ns;
 */
  long					hilightwait=100000000L;


 


  public LCDControler(LCDDisplayGUI display) {
    super(8, 4);

    this.display=display;
    initContrastFSM();
    initHilightFSM();
//    initAddrFSM();
  }


  void initContrastFSM() {
    StateDef		state;
    String		group="Contrast";
    ActionMethod	mon, moff;

    try {
      Class prms[]={Boolean.TYPE};
      Object objs[];

      objs=new Object[1];	objs[0]=new Boolean(true);
      moff=new ActionMethod(display, display.getClass().getMethod("setContrast",
                            prms), objs);

      objs=new Object[1];	objs[0]=new Boolean(false);
      mon=new ActionMethod(display, display.getClass().getMethod("setContrast",
                           prms), objs);

      initPinSchmittTrigger(group, spike, contrwait, VARS_CONTRAST,
                            VARS_CONTRAST, moff, mon, false);

    } catch (NoSuchMethodException nme) {
      throw new InternalError();
    }
    setInitialState("Contrast InitLow", group);
  }



  void initHilightFSM() {
    StateDef            state;
    String              group="Hilight";
    ActionMethod        mon, moff;
 
    try {
      Class prms[]={Boolean.TYPE};
      Object objs[];
 
      objs=new Object[1];       objs[0]=new Boolean(false);
      moff=new ActionMethod(display, display.getClass().getMethod("setHilight",
                            prms), objs);
 
      objs=new Object[1];       objs[0]=new Boolean(true);
      mon=new ActionMethod(display, display.getClass().getMethod("setHilight",
                           prms), objs);
 
      initPinSchmittTrigger(group, spike, contrwait, VARS_HILIGHT,
                            VARS_HILIGHT, moff, mon, false);
 
    } catch (NoSuchMethodException nme) {
      throw new InternalError();
    }
    setInitialState(group+" InitLow", group);
  }





/** Init the final state machine for the address lines rs and
 *  rw. Both lines are have sample and hold mode and will print
 *  error messages when changes during hold occur.
 */
  void initAddrFSM() {
    try {
      initPinSchmittTrigger("RW", spike, -1, VARS_RW, TIMERS_RW,
                            null, null, true);
      setInitialState("RW InitLow", "RW");
    } catch (NoSuchMethodException nsm) {
      nsm.printStackTrace();
      throw new InternalError();
    }
  }






/** Create a schmitt trigger on one pin. The trigger knows
 *  following states:<BR>
 *  <Ul><Li>InitLow: set counter 0 -> Low
 *  <Li>Low: if high -> InitHigh, if unknown -> WaitRise
 *  <Li>WaitRise: high -> InitHigh, low -> Low, unknown + timer -> Rise
 *  <Li>Rise: high -> initHigh, low -> Low+Error, counter -> RiseError
 *  </Ul>
 *  @param spike max time for an invalid signal not triggering
 *  @param maxrise max time until signal reaches other level
 *  @param vpos position of the first variable
 *  @param tpos position of the timer variable
 *  @param hold if true the variable vpos+1 is a sample and hold
 *  trigger (0 samples data, 1 holds it), vpos+2 the sampled
 *  value of the schmitt trigger element
 */
  void initPinSchmittTrigger(String name, long spike, long maxrise, int vpos,
                             int tpos, ActionMethod offmeth,
                             ActionMethod onmeth, boolean hold)
                             throws NoSuchMethodException {

    StateDef			st_ilow, st_low, st_ihigh, st_high;
    java.lang.reflect.Method	sample_method;
    Class			prms[]=null;
    Transition			t, gethigh, getlow;
    Object			args[];

    setValue(Pin.HiZ, vpos);
    createStateGroup(name);

// The pin low init, sets timer back to 0
// if hold function requested, write 0 to vpos+2 if vpos+1=0
    st_ilow=new StateDef(this, name+" InitLow", name);



// add pin low state, decrease timer while staying
// on unknown go to WaitRise
// on high go to init high
    st_low=new StateDef(this, name+" Low", name);
    st_low.stayactions=new StateAction[1];
    st_low.stayactions[0]=new StateAction(StateAction.ACTION_DECTIMER, tpos,
                                          -1, -1, 0);


// The pin high init, sets timer back to 0
// if hold function requested, write 0 to vpos+2 if vpos+1=0
    st_ihigh=new StateDef(this, name+" InitHigh", name);


// add pin high state, decrease timer while staying
// on unknown go to WaitFall
// on low go to init low
    st_high=new StateDef(this, name+" High", name);
    st_high.stayactions=new StateAction[1];
    st_high.stayactions[0]=new StateAction(StateAction.ACTION_DECTIMER, tpos,
                                           -1, -1, 0);


// if sample/hold functionality requested, init the actionmethod
// now, attatch the on and offmethods to call them from inside
// the s/h method
    sample_method=null;
    if(hold) {
      prms=new Class[3];
      prms[0]=Integer.TYPE;
      prms[1]=String.class;
      prms[2]=ActionMethod.class;
      sample_method=getClass().getMethod("doSample", prms);
    }


//
// now add the states transitions
//
    t=new Transition(this);
    t.changeactions=new StateAction[1];
    t.changeactions[0]= new StateAction(StateAction.ACTION_SETTIMER, tpos,
                                        -1, -1, 0);

    if(hold) {
// do sampling
      args=new Object[3];
      args[0]=new Integer(vpos);
      args[1]=name;
      args[2]=offmeth;
      t.changemethod=new ActionMethod(this, sample_method, args);
    } else {
// unconditional transition
      t.changemethod=offmeth;
    }
    st_ilow.addTransition(t, st_low);


    gethigh=new Transition(this);
    gethigh.conditions=new int[1][];
    gethigh.conditions[0]=createIsHighCondition(vpos);
    st_low.addTransition(gethigh, st_ihigh);


    t=new Transition(this);
    t.changeactions=new StateAction[1];
    t.changeactions[0]= new StateAction(StateAction.ACTION_SETTIMER, tpos,
                                        -1, -1, 0);
    if(hold) {
// do sampling
      args=new Object[3];
      args[0]=new Integer(vpos);
      args[1]=name;
      args[2]=onmeth;
      t.changemethod=new ActionMethod(this, sample_method, args);
    } else {
// unconditional transition
      t.changemethod=offmeth;
    }
    st_ihigh.addTransition(t, st_high);
 
 
    getlow=new Transition(this);
    getlow.conditions=new int[1][];
    getlow.conditions[0]=createIsLowCondition(vpos);
    st_high.addTransition(getlow, st_ilow);
  }





/** Sample if the value at vpos+1 is low.
 */
  public void doSample(int vpos, String group, ActionMethod action) {
    if(statevars[vpos+1]==0) newvars[vpos+2]=statevars[vpos];
    else { // val change during hold
      println(group+"-pin didn't keep state, switched to "+statevars[vpos]);
    }
    if(action!=null) action.invoke();
  }




  int[] createIsHighCondition(int varnum) {
    int r[]=new int[4];

    r[0]=Transition.COND_ISIN;
    r[1]=varnum;
    r[2]=Pin.HIGH;
    r[3]=Pin.WEAKHIGH;
    return(r);
  }

  int[] createIsLowCondition(int varnum) {
    int r[]=new int[4];
 
    r[0]=Transition.COND_ISIN;
    r[1]=varnum;
    r[2]=Pin.LOW;
    r[3]=Pin.WEAKLOW;
    return(r);
  }


  int[] createIsUnknownCondition(int varnum) {
    int r[]=new int[5];

    r[0]=Transition.COND_ISIN;
    r[1]=varnum;
    r[2]=Pin.UNKNOWN;
    r[3]=Pin.WEAKUNKN;
    r[4]=Pin.HiZ;
    return(r);
  }



// ************************************************************************* //
//									     //
//			Glue to LCDDisplay				     //
//									     //
// ************************************************************************* //


  public void setValues(int values[]) {
    //ARM (".setValues()");
    setValue(values[LCDDisplay2.PIN_CONTRAST], VARS_CONTRAST);
    setValue(values[LCDDisplay2.PIN_HILIGHT], VARS_HILIGHT);
  }


  public void println(String text) {
    //ERROR ("#: msg: "+text);
  }


}






