package fips.sim.dig.comp;

import java.awt.Graphics;

/** Simulation of the H44787 LCD Display + controller
with 5x8 and 5x10 fonts, 2 pix horiz space, 1pix vert to cursor,
2 (5x10) or 4 (5x8) pix more to next row

So component is at least 2 units high and 8 long for 1x8 display,
3 high and 16 long for 2x16 disp


8 5x8 characters are definable
4 5x10

Display works only if scale+scalef >= 4 for 1 pix per dot


Pins and positions are:
D0 .. D7 upper border
EN, RW, RS, Vo, Hilight lower border



What sequences change between cgram, ddram access?


 *  Read behavior: After a set address instruction content from
 *  ram is moved to DR, then the BF is deasserted. The next read
 *  will return the data from DR immediatly, assert the BF and
 *  fetch the next data if increment mode active. After BF gets
 *  low, another 1.5 cycles are needed until AC is updated.
 */



public class LCDDisplay implements fips.sim.dig.base.Component {

  String			name;

  public final static int	PIN_DATA0	=  0;
  public final static int	PIN_DATA7	=  7;
  public final static int	PIN_ENABLE	=  8;
  public final static int	PIN_RW		=  9;
  public final static int	PIN_RS		= 10;
  public final static int	PIN_CONTRAST	= 11;
  public final static int	PIN_HILIGHT	= 12;

  fips.sim.dig.comp.Pin		pins[];
  int				pinnums[];
  int				states[];

/** State of each pin before current state was reached
 */
  int				laststates[];

/** Time elapsed since pin's state change from laststate to current
    state. This value can't get bigger than maxvalue
 */
  long				lastchgs[];



/**Scalefactor to allow bigger display
*/
  int				scalef;

  LCDDisplayGUI			lcdgui;



// Component specific variables


/** Current cursor pos in ddram
*/
  int				ramaddr;


/** If true a read or write command increments ram address
 */
  boolean			incaddrmode;


/** If true ram rws affect the display data ram.
 */
  boolean			ddrammode;
/** If true ram rws affect the character generator ram.
 */
  boolean			cgrammode;


/** State of the busy flag in the controller. After a read or
 *  write ram instruction, the flag may be off but the busytimer
 *  set while controller changes ddram address value
 */
  boolean			busyflag;


/** State of the last command executed. This is needed to allow
 *  additional busy cycles with busyflag false after an instruction
 *  increasing the address register.
 */
  int				lastaction;


/** State of the function set lines at the begin of an instruction.
 *  The state is reduced to -1, 0, 1 for unknown, low and high.
 */
  int				rwline;
  int				rsline;



/** If true any exceeding of maximum/minimum timing specifications
    will result in a warning message. Also an unknown signal
    that is longer than a spike will result in a message.
 */
  boolean			strict;



  final static int		STATE_ENLOW		= 0;
  final static int		STATE_ENLOWTOHIGH	= 1;

// time to put data onto bus in read mode, while lines are unknown
  final static int		STATE_RDRIVEDATA	= 2;

// read data fully driven onto bus
  final static int		STATE_RDATAREADY	= 3;

// en going down, this state is entered spike fs after en
// leaves high
  final static int		STATE_RHIGHTOLOW	= 4;

// data hold while en low
  final static int		STATE_RDATALOW		= 5;

// address hold while en low
  final static int		STATE_RADDRHOLD		= 6;


// state if minimum cycle time still hasn't elapsed
  final static int		STATE_FINISHCYCLE	= 0xF;




// uninitialized, EN low
  final static int		STATE_UNINIT	= 0;

// en low, ready to await function set or instruction
  final static int		STATE_RESET	= 0x100;

// en low, doing nothing
  final static int		STATE_IDLE	= 0x110;

  final static int		STATE_READRAM	= 0x200;
  final static int		STATE_WRITERAM	= 0x210;

  final static int		STATE_MASK	= 0xFFF0;



/** Current controller state of the lcd panel
 */
  int				lcdstate;



// timing critical values


/** Absolute running time of this device
 */
  java.math.BigInteger		time;



/** Maximum length of an unknown signal on a pin until a state
    may get triggered. The default value is 1 ns. 
 */
  long                          spike=1000000;                                  


/** Internal clock period, currently 4us (250kHz)
 */
  long				intclock=4000000000L;


/** Time for a reset performance, max 10ms
 */
  long				resettime=10000000000000L;

/** If true, the unit performs a normal reset, no init by
    instruction needed
 */
  boolean			normalreset;



/** Time for a pinchange on the contrast pin to show effects.
    The default value is 100ns;
 */
  long				contrwait=100000000L;


/** Accumulated time after contrast pin change. If contrast
    is off, stepping increases value until contrwait reached.
    If pin then is not 1 error is given (unknown state). If
    contrast is on, contracc is decreased until 0 reached, then
    lcd driver is switched off.
 */
  long				contracc;
  int				contrlaststate;


/** Time for a pinchange on the hilight pin to show effects.
 *  The default value is 100ns;
 */
  long				hilightwait=100000000L;
 
 
/** Accumulated time after hilight pin change. If hilight
 *  is off, stepping increases value until hilightwait reached.
 *  If pin then is not 1 error is given (unknown state). If
 *  hilight is on, hilightacc is decreased until 0 reached, then
 *  lcd led driver is switched off.
 */
  long				hilightacc;
  int				hilightlaststate;




/** Minimum valid cycle time from one to next rise of enable
    pin: 500ns at 4.5V, 1000ns at 2.3V
 */
  long				tcycmin=500000000L;

  java.math.BigInteger		cycstart;


/** Max time for an enable rise or fall, 20ns at 4.5V, 25ns
    at 2.3 V
 */
  long				enableedgemax=20000000L;


/** Cycles since start of enable edge
 */
  long				enableedgetime;


/** Minimum pulse width of enable signal: 230ns at 4.5V, 450 at 2.3V
 */
  long				tenablepw=230000000L;

/** Time of the last enable pin transition to valid high or
    low state.
 */
//  java.math.BigInteger		enpstart;
  long				enplength;



/** Time before rise of enable that the address lines have
 *  to be valid: 40ns at 4.5V, 60ns at 2.3V
 */
  long				taddrset=40000000L;


/** Time after falling edge of en the address lines have to stay
 *  valid: 10ns at 4.5V, 20ns at 2.3V
 */
  long				taddrhold=10000000L;


/** Time from begin of read instruction (en high) until bus data
 *  is valid: 160ns at 4.5V, 360 at 2.3V
 */
  long				tdataready=160000000L;                          


/** Time read data stays ready after en got low again:
 *  5ns at 4.5V, 10ns at 2.3V
 */
  long				tdatahold=5000000L;


/** Time the controller will be busy. Execution of commands other than
 *  read address (rw=1, rs=0) will result in an error.
 */
  long				busytime;





  public LCDDisplay() {
    int i;

    name=getComponentName();

// initialize the pins
    pins=new fips.sim.dig.comp.Pin[13];
    pinnums=new int[13];
    states=new int[13];
    laststates=new int[13];
    lastchgs=new long[13];

    for(i=0; i<states.length; i++) {
      states[i]=Pin.HiZ;
      laststates[i]=Pin.UNKNOWN;
      lastchgs[i]=0;
    }


    lcdgui=new LCDDisplayGUI();

// set preferred scale, scalefactor and real diplay type
    lcdgui.setType(4, 1, lcdgui.TYPE_SMALL2x16);


    normalreset=true;
    lcdstate=(normalreset?STATE_RESET:STATE_UNINIT);

    lcdgui.setHilight(false);
    lcdgui.setContrast(false);

    time=java.math.BigInteger.ZERO;

    contracc=0;
    contrlaststate=Pin.HiZ;

    hilightacc=0;
    hilightlaststate=Pin.HiZ;

// increment ramaddr ?
    incaddrmode=true;

    cgrammode=false;
    ddrammode=false;


    cycstart=time;

    busytime=resettime;
    busyflag=false;

    lastaction=0;
  }




  public String getComponentName() {
    return("LCDDisplay");
  }




  public String getDescription() {
    return("LCD Display");
  }


  public String getPinDescription(int num) {
    throw new InternalError();
  }


  public String getName() {
    return(name);
  }
 
 
  public void setName(String name) {
    this.name=name;
  }


  public int getParameterCount() {
    throw new InternalError();
  }

  public fips.sim.dig.base.Parameter[] getParameters() {
    throw new InternalError();
  }


  public void setParameter(int num, Object val) {
    throw new InternalError();
  }



  public int getPinCount() {
    return(13);
  }




  public void setPin(int num, fips.sim.dig.comp.Pin p) {
    if(num>=13) throw new IllegalArgumentException("Can't set pin "+num);

    if(pins[num]!=null) pins[num].removeComponent(pinnums[num]);

    pins[num]=p;
    if(pins[num]!=null) {
      pinnums[num]=pins[num].addComponent();
      pins[num].setValue(Pin.HiZ, pinnums[num]);
    }
  }






  public fips.sim.dig.comp.Pin getPin(int num) {
    if(num>=13) throw new IllegalArgumentException();
    return(pins[num]);
  }








/** Get a hint for max length of next step. This method just
    calculates the step length and sets variables due to pin
    changes, but won't alter any state or pin itself.
 */
  public long getNextStep() {
    int				i, j;
    long			l, nextstep;
    java.math.BigInteger	bi;

    for(i=0; i<pins.length; i++) {
      j=pins[i].getValue();
      if(j!=states[i]) {
        switch(j) {
          case Pin.LOW:
          case Pin.WEAKLOW:
            if(isPinLow(i)) {
              states[i]=j;	// no real change
            }
            break;
          case Pin.HIGH:
          case Pin.WEAKHIGH:
            if(isPinHigh(i)) {
              states[i]=j;      // no real change
            }
            break;
          case Pin.UNKNOWN:
          case Pin.WEAKUNKN:
          case Pin.HiZ:
            if(isPinUnknown(i)) {
              states[i]=j;      // no real change
            } 
            break;
          default: throw new InternalError();
        }

        if(j!=states[i]) {
          laststates[i]=states[i];
          states[i]=j;
          lastchgs[i]=0;
        }
      }
    }



    nextstep=Long.MAX_VALUE;


// check state of contrast pin changed
    if(states[PIN_CONTRAST]!=contrlaststate) {
      switch(contrlaststate) {
        case Pin.LOW:
        case Pin.WEAKLOW:
        case Pin.HiZ:
          if((isPinLow(PIN_CONTRAST))||
             (states[PIN_CONTRAST]==Pin.HiZ)) {
            contrlaststate=states[PIN_CONTRAST];
            break;
          }

// pin changed, calc nextstep from contrwait
          l=contrwait-contracc;
          if(l<nextstep) nextstep=l;
          break;

        case Pin.HIGH:
        case Pin.WEAKHIGH:
          if(isPinHigh(PIN_CONTRAST)) {
            contrlaststate=states[PIN_CONTRAST];
            break;
          }

// pin changed, decrease contracc
          if(contracc<nextstep) nextstep=contracc;
          break;

        default:
          throw new InternalError("Illegal state "+contrlaststate);
      }
    }




// check state of hilight pin changed
    if(states[PIN_HILIGHT]!=hilightlaststate) {
      switch(hilightlaststate) {
        case Pin.LOW:
        case Pin.WEAKLOW:
        case Pin.HiZ:
          if((isPinLow(PIN_HILIGHT))||
             (states[PIN_HILIGHT]==Pin.HiZ)) {
            hilightlaststate=states[PIN_HILIGHT];
            break;
          }

// pin changed, calc nextstep from hilightwait
          l=hilightwait-hilightacc;
          if(l<nextstep) nextstep=l;
          break;

        case Pin.HIGH:
        case Pin.WEAKHIGH:
          if(isPinHigh(PIN_HILIGHT)) {
            hilightlaststate=states[PIN_HILIGHT];
            break;
          }

// pin changed, decrease hilightacc
          if(hilightacc<nextstep) nextstep=hilightacc;
          break;

        default:
          throw new InternalError("Illegal state "+hilightlaststate);
      }
    }




    switch(lcdstate&0xF) {
      case STATE_ENLOW:	// waiting for instructions
        if((states[PIN_ENABLE]==Pin.LOW)||(states[PIN_ENABLE]==Pin.WEAKLOW)) {
          break;
        }

// count cyclestart from begin of rise
        cycstart=time;
        enableedgetime=0;

        if(isPinHigh(PIN_ENABLE)) {
// directly enter high state 
          nextstep=0;
        }

        lcdstate|=STATE_ENLOWTOHIGH;
        if(nextstep>spike) nextstep=spike;

        break;
      case STATE_RDRIVEDATA:
        if(enplength>=tdataready) {
          throw new InternalError("Can't be in this state for such long time.");
         }

        if(isPinLow(PIN_ENABLE)) { // illegal low transition of en
          nextstep=0;
        } else {
          l=tdataready-enplength;
          if(l<nextstep) nextstep=l;
        }
 
        break;

      case STATE_RDATAREADY:
// this state is stable, so display may wait indefinitly if
// enable stays high and rws doesn't change
        if((isPinLow(PIN_ENABLE))||(getReducedPinState(PIN_RW)!=rwline)||
           (getReducedPinState(PIN_RS)!=rsline)) {
          nextstep=0;
          break;
        }
        if(isPinUnknown(PIN_ENABLE)) {
          l=spike-lastchgs[PIN_ENABLE];
          if(nextstep>l) nextstep=l;
        }
        break;

      case STATE_RDATALOW:
// read complete, wait indefinitly if en and rws stable

        if((isPinHigh(PIN_ENABLE))||(getReducedPinState(PIN_RW)!=rwline)||
           (getReducedPinState(PIN_RS)!=rsline)) {
          nextstep=0;
        } else {
          l=tdatahold-enplength;
          if(l<nextstep) nextstep=l;

          if((isPinUnknown(PIN_ENABLE))&&(nextstep>spike)) nextstep=spike;
        }
        break;

      case STATE_RADDRHOLD:
        if((isPinHigh(PIN_ENABLE))||(getReducedPinState(PIN_RW)!=rwline)||
           (getReducedPinState(PIN_RS)!=rsline)) {
          nextstep=0;
        } else {
          l=taddrhold-enplength;
          if(l<nextstep) nextstep=l;

          if((isPinUnknown(PIN_ENABLE))&&(nextstep>spike)) nextstep=spike;
        }
        break;

      case STATE_FINISHCYCLE:
        if(isPinHigh(PIN_ENABLE)) {
          nextstep=0;
          break;
        }

        if(isPinUnknown(PIN_ENABLE)) {
          l=spike-lastchgs[PIN_ENABLE];
          if(l<nextstep) nextstep=l;
        }

        if((bi=time.subtract(cycstart)).bitLength()<64) {
          l=tcycmin-bi.longValue();
          if(l<nextstep) nextstep=l;
        } 
        break;

      default:
        throw new InternalError("Invalid state 0x"+
                                 Integer.toHexString(lcdstate));
    }

    if((busytime>0)&&(busytime<nextstep)) nextstep=busytime;

    return(nextstep);
  }









/** Do the following calculation step. If cycles is too big,
    IllegalArgumentException is thrown
 */
  public void doStep(long cycles) {
    int				i, n;
    java.math.BigInteger	bi;


    for(i=0; i<lastchgs.length; i++) {
      lastchgs[i]+=cycles;
      if(lastchgs[i]<0) lastchgs[i]=Long.MAX_VALUE;	// overflow
    }

    if(cycles>0) time=time.add(java.math.BigInteger.valueOf(cycles));




// do the contrast pin
// if contrast was off, contracc is increased
    if(contrlaststate!=states[PIN_CONTRAST]) {
// getNextStep should have inhibited too long steps
      if(cycles>contrwait) throw new InternalError();

      contracc+=(lcdgui.getContrast()?-cycles:cycles);

      if(lcdgui.getContrast()) {
// acc can't get <0
        if(contracc<0) throw new InternalError();

        if(contracc==0) {
// if signal now valid switch contrast state
          switch(states[PIN_CONTRAST]) {
            case Pin.UNKNOWN:
            case Pin.WEAKUNKN:
              System.err.println("Contrast pin of "+getName()+
                                 " didn't reach valid state in time.");
            case Pin.LOW:
            case Pin.WEAKLOW:
            case Pin.HiZ:
              //TRACE ("#: disabling contrast");
              contrlaststate=Pin.LOW;
              lcdgui.setContrast(false);
              break;
 
            default:
// pin can't be low, cause contracc is only increased when
// pin != low
              throw new InternalError("Illegal state "+states[PIN_CONTRAST]);
          }
        }

      } else {
// contrast now switched on

        if(contracc>contrwait) throw new InternalError();

        if(contracc==contrwait) {
// if signal now valid switch contrast state
          switch(states[PIN_CONTRAST]) {
            case Pin.UNKNOWN:
            case Pin.WEAKUNKN:
              System.err.println("Contrast pin of "+getName()+
                                 " didn't reach valid state in time.");
            case Pin.HIGH:
            case Pin.WEAKHIGH:
              //TRACE ("#: enabling contrast");
              contrlaststate=Pin.HIGH;
              lcdgui.setContrast(true);
              break;

            default:
// pin can't be low, cause contracc is only increased when
// pin != low
              throw new InternalError("Illegal state "+states[PIN_CONTRAST]);
          }
        }
      }		// end of if(contrast)
    } else {
// contrastpin has last state, so acc will go to 0 for contrast off
      if(lcdgui.getContrast()) {
        contracc+=cycles;
        if((contracc>contrwait)||(contracc<0)) contracc=contrwait;
      } else {
        contracc-=cycles;
        if(contracc<0) contracc=0;
      }
    }





// do the hilight pin
// if hilight was off, hilightacc is increased
    if(hilightlaststate!=states[PIN_HILIGHT]) {
// getNextStep should have inhibited too long steps
      if(cycles>hilightwait) throw new InternalError();

      hilightacc+=(lcdgui.getHilight()?-cycles:cycles);

      if(lcdgui.getHilight()) {
// acc can't get <0
        if(hilightacc<0) throw new InternalError();

        if(hilightacc==0) {
// if signal now valid switch hilight state
          switch(states[PIN_HILIGHT]) {
            case Pin.UNKNOWN:
            case Pin.WEAKUNKN:
              System.err.println("Hilight pin of "+getName()+
                                 " didn't reach valid state in time.");
            case Pin.LOW:
            case Pin.WEAKLOW:
            case Pin.HiZ:
              //TRACE ("#: switching off background light");
              hilightlaststate=Pin.LOW;
              lcdgui.setHilight(false);
              break;
 
            default:
// pin can't be low, cause hilightacc is only increased when
// pin != low
              throw new InternalError("Illegal state "+states[PIN_HILIGHT]);
          }
        }

      } else {
// hilight now switched on

        if(hilightacc>hilightwait) throw new InternalError();

        if(hilightacc==hilightwait) {
// if signal now valid switch hilight state
          switch(states[PIN_HILIGHT]) {
            case Pin.UNKNOWN:
            case Pin.WEAKUNKN:
              System.err.println("Hilight pin of "+getName()+
                                 " didn't reach valid state in time.");
            case Pin.HIGH:
            case Pin.WEAKHIGH:
              //TRACE ("#: Enabling background light of "+getName());
              hilightlaststate=Pin.HIGH;
              lcdgui.setHilight(true);
              break;

            default:
// pin can't be low, cause hilightacc is only increased when
// pin != low
              throw new InternalError("Illegal state "+states[PIN_HILIGHT]);
          }
        }
      }		// end of if(hilightt)
    } else {
// hilightpin has last state, so acc will go to 0 for hilight off
      if(lcdgui.getHilight()) {
        hilightacc+=cycles;
        if((hilightacc>hilightwait)||(hilightacc<0)) hilightacc=hilightwait;
      } else {
        hilightacc-=cycles;
        if(hilightacc<0) hilightacc=0;
      }
    }







    if(busytime>0) {
      busytime-=cycles;

      if(busytime<0) throw new InternalError();

      if(busytime==0) {
        busytime=0;
        busyflag=false;

        //TRACE ("#: leaving busy state in component "+name);

        if((lastaction==STATE_READRAM)||(lastaction==STATE_WRITERAM)) {
// if the last action was a rw ram command, the address is increased
// afterwards, which needs 1.5 internal clock cycles of the controller

          if(incaddrmode) ramaddr++;
          else ramaddr--;
          ramaddr&=0x7F;

          busytime=intclock+(intclock>>1);
          lastaction=STATE_IDLE;
        }
      }
    }



// do the state transitions in the controller

    switch(lcdstate&0xF) {
      case STATE_ENLOW:	// stay low
        break;
      case STATE_ENLOWTOHIGH:
        enableedgetime+=cycles;
        if(enableedgetime<0) enableedgetime=Long.MAX_VALUE;

        if(isPinHigh(PIN_ENABLE)) { // enter high state
          if((strict)&&(enableedgetime>=enableedgemax)) {
            System.out.println("Rise of enable of "+getName()+
                               " not fast enough, check circuit design.");
          }

// check the address lines for next state
          enplength=0;
 
          rwline=getReducedPinState(PIN_RW);
          rsline=getReducedPinState(PIN_RS);
 
          if(rwline==-1) {
            System.err.println("RW-pin of "+getName()+
                               " has unknown state, check circuit design.");
          } else {
            if((lastchgs[PIN_RW]<taddrset)&&(strict)) {
              System.out.println("RW-pin not stable long enough");
            }
          } 
 
          if(rsline==-1) {
            System.err.println("RS-pin of "+getName()+
                               " has unknown state, check circuit design.");
          } else {
            if((lastchgs[PIN_RS]<taddrset)&&(strict)) {
              System.out.println("RS-pin not stable long enough");
            }
          }

 
          if(rwline==1) {
// starting a read cycle, a read during init will result in an
// error message and termination of the current init process
            if((lcdstate&STATE_MASK)<STATE_IDLE) {
              System.err.println("Read on uninitialized device "+getName()+
                                 ", check circuit design.");

              busytime=0;
              lcdstate=STATE_IDLE|STATE_ENLOWTOHIGH;
              busyflag=false;
            }

            if(rsline==1) {		// read ram data
              if((busyflag)||(busytime>0)) {
                System.err.println("Read ram call on device "+getName()+
                                   " while busy, check design.");
                busyflag=false;
                busytime=0;
              }


              if((!ddrammode)&&(!cgrammode)) {
                System.err.println("Trying to read data while source (cgram or ddram) is not clearly specified!");
                System.err.println("You may get this warning because the datasheet is not accurate on how\ntransitions between the two states work, so I assume a ddram read.");
                ddrammode=true;
              }
            }

            for(i=0; i<8; i++) pins[i].setValue(Pin.UNKNOWN, pinnums[i]);
            lcdstate=STATE_READRAM|STATE_RDRIVEDATA;

            break;
          }


// start write cycle

          if(rsline==1) { // write data
          }

// write instruction

          throw new InternalError("Write not implemented");

//          break;
        }




        if(isPinLow(PIN_ENABLE)) {
// get low again,
          lcdstate=(lcdstate&STATE_MASK)|STATE_ENLOW;
          if((enableedgetime>=spike)&&(strict)) {
            System.out.println("Enable got unknown for more than spike time, check your circuit design.");
          }
        }

// unknown states are not considered, due to fact that they have
// to get known again some time, then the error message will occur
        break;


      case STATE_RDRIVEDATA:
        enplength+=cycles;
        if(enplength<0) enplength=Long.MAX_VALUE;

// if en is low give error message
        if(isPinLow(PIN_ENABLE)) {
          System.err.println("Enable pin got low too fast, terminating read cycle.\nCheck your circuit design.");
          lcdstate=STATE_IDLE;
          break;
        }

// check if address data is still valid
        if((getReducedPinState(PIN_RW)!=rwline)||
           (getReducedPinState(PIN_RS)!=rsline)) {
          System.err.println("Addresslines of "+getName()+
                             " not stable while chip enabled, check circuit design.");
        }


        if(enplength>tdataready) throw new InternalError();

// data still not valid, continue
        if(enplength<tdataready) break;

// enter next state (STATE_RDATAREADY) by driving data onto bus
        lcdstate++;

        i=-1;
        if(rsline==1) {
          // ram read
          if(ddrammode) i=lcdgui.getRamData(ramaddr);
          if(cgrammode) throw new InternalError();
        } else {
          i=ramaddr|(busyflag?0x80:0);
        }

        if(i==-1) throw new InternalError();

        //TRACE ("#: writing value 0x"+Integer.toHexString(i)+
        //TRACE  " to data port of component "+getName());
        for(n=0; n<8; n++) {
          pins[n].setValue((i&(1<<n))!=0 ? Pin.HIGH : Pin.LOW, pinnums[n]);
        }
        break;

      case STATE_RDATAREADY:
        enplength+=cycles;
        if(enplength<0) enplength=Long.MAX_VALUE;

// check if address data is still valid
        if((getReducedPinState(PIN_RW)!=rwline)||
           (getReducedPinState(PIN_RS)!=rsline)) {
          System.err.println("Addresslines of "+getName()+
                             " not stable while chip enabled, check circuit design.");
        }

        if(isPinHigh(PIN_ENABLE)) break;	// no change

// enter next state, check if pulse width was ok
        if(isPinLow(PIN_ENABLE)) {
          if(cycles!=0) {
            throw new InternalError("getStep failed, cycles="+cycles+" found.");
          }

// check if pulse was long enough
          if(enplength<tenablepw) {
            System.err.println("Enable pulse for device "+getName()+
                               " too short, check circuit design.");
          }

          lcdstate+=2;
          enplength=0;		// use enplength to measure time since en-edge

          //TRACE ("#: entering read data low state in "+name);

// if reading from ram set busytime
          if(rsline==1) {
            busytime=intclock*10L;
            busyflag=true;
          }
          break;
        }

// enable pin unknown
        if(lastchgs[PIN_ENABLE] > spike) throw new InternalError();
        if(lastchgs[PIN_ENABLE] < spike) break;

        lcdstate++;
        //TRACE ("#: entering read high to low state in "+name);
        break;

      case STATE_RDATALOW:
        enplength+=cycles;
        if(enplength<0) enplength=Long.MAX_VALUE;        

// check if address data is still valid
        if((getReducedPinState(PIN_RW)!=rwline)||
           (getReducedPinState(PIN_RS)!=rsline)) {
          System.err.println("Addresslines of "+getName()+
                             " not stable while chip enabled, check circuit design.");
        }

        if((isPinHigh(PIN_ENABLE))||
           ((isPinUnknown(PIN_ENABLE))&&(lastchgs[PIN_ENABLE]==spike))) {
          System.err.println("Enable on device "+getName()+
                             " started to rise while cycle not complete, check circuit design.");
          throw new InternalError();
        }

        if(enplength==tdatahold) {
// tristate pins
          for(i=0; i<8; i++) pins[i].setValue(Pin.HiZ, pinnums[i]);
          lcdstate++;
          //TRACE ("#: entering read address hold state in "+name);
        }
        break;

      case STATE_RADDRHOLD:
        enplength+=cycles;
        if(enplength<0) enplength=Long.MAX_VALUE;

// check if address data is still valid
        if((getReducedPinState(PIN_RW)!=rwline)||
           (getReducedPinState(PIN_RS)!=rsline)) {
          System.err.println("Addresslines of "+getName()+
                             " not stable while chip enabled, check circuit design.");
        }

        if((isPinHigh(PIN_ENABLE))||
           ((isPinUnknown(PIN_ENABLE))&&(lastchgs[PIN_ENABLE]==spike))) {
          System.err.println("Enable on device "+getName()+
                             " started to rise while cycle not complete, check circuit design.");
          throw new InternalError();
        }

        if(enplength==taddrhold) {
// all ops complete enter finish-cycle state
          lcdstate&=STATE_MASK;

          if(((bi=time.subtract(cycstart)).bitLength()<64)&&
              (bi.longValue()<tcycmin)) {
            lcdstate|=STATE_FINISHCYCLE;
            //TRACE ("#: entering finish state in "+name);
          } else {
            //TRACE ("#: entering idle state in "+name);
          }
        }
        break;

      case STATE_FINISHCYCLE:
        enplength+=cycles;
        if(enplength<0) enplength=Long.MAX_VALUE;

        if(isPinHigh(PIN_ENABLE)) {
          System.err.println("Time between 2 enable cycles for component "+
                             getName()+" too short, check circuit design.");
          throw new InternalError();
        }

        if((isPinUnknown(PIN_ENABLE))&&(lastchgs[PIN_ENABLE]==spike)) {
          System.err.println("Slow enable pin rise for component "+name+
                             " while last enable cycle not complete.");
        }

        if(((bi=time.subtract(cycstart)).bitLength()>=64)||
           (bi.longValue()>=tcycmin)) {
          lastaction=lcdstate&STATE_MASK;
          lcdstate=STATE_IDLE;
          //TRACE ("#: entering idle state in "+name);
        }
        break;


      default:
        throw new InternalError("Illegal lcd state 0x"+
                                Integer.toHexString(lcdstate));
    }


  }









/** Reset cyclecount, params, etc...
 */
  public void reset() {
    throw new InternalError();
  }


  public void writeState(java.io.ObjectOutputStream oos)
                         throws java.io.IOException {
    throw new InternalError();
  }
 
 
  public void readState(java.io.ObjectInputStream ois)
                        throws java.io.IOException {
    throw new InternalError();
  }



/**Get the GUI size of this component.
*/
  public java.awt.Dimension getSize() {
    return(lcdgui.getSize());
  }



  public void setGUIParameters(java.awt.Point offs, int scale, int rot) {
    throw new InternalError();
  }


  public int getOrientation() {
    throw new InternalError();
  }


/** Init the graphics with given scale and calculate the values
    of boarders and spaces. 
*/
  public void repaint(Graphics g) {
  }
  public void update(Graphics g) {
  }

  public boolean needsUpdate() { return(true); }


/** Take actions if clicked, currently none
*/
  public void doClick(int x, int y, int button) {
  }



  public java.awt.Point getPinPos(int num) {
    throw new InternalError();
  }




// ************************************************************************* //
//
//			Internal Methods				     //
//
// ************************************************************************* //











// ************************************************************************* //
//
//			Display state changes				     //
//
// ************************************************************************* //






  boolean isPinHigh(int num) {
    return((states[num]==Pin.HIGH)||(states[num]==Pin.WEAKHIGH));
  }

  boolean isPinLow(int num) {
    return((states[num]==Pin.LOW)||(states[num]==Pin.WEAKLOW));
  }

  boolean isPinUnknown(int num) {
    return((states[num]==Pin.UNKNOWN)||(states[num]==Pin.WEAKUNKN)||
           (states[num]==Pin.HiZ));
  }


/** Return the simplified pin state for a given pin. If the
 *  pin value is driven hard or weak, the state is returned, if
 *  undecidable the result is -1
 */
  int getReducedPinState(int num) {
    if((states[num]==Pin.HIGH)||(states[num]==Pin.WEAKHIGH)) return(1);
    if((states[num]==Pin.LOW)||(states[num]==Pin.WEAKLOW)) return(0);
    return(-1);
  }



  public Object clone() {
    throw new InternalError();
  }


}






