package fips.sim.dig.comp;

import java.awt.Graphics;

import fips.sim.dig.comp.fsm.*;

/**Simulation of the H44787 LCD Display + controller
 * with 5x8 and 5x10 fonts, 2 pix horiz space, 1pix vert to cursor,
 * 2 (5x10) or 4 (5x8) pix more to next row

 * So component is at least 2 units high and 8 long for 1x8 display,
 * 3 high and 16 long for 2x16 disp


 * 8 5x8 characters are definable
 * 4 5x10

 * Display works only if scale+scalef >= 4 for 1 pix per dot


 * Pins and positions are:
 * D0 .. D7 upper border
 * EN, RW, RS, Vo, Hilight lower border



 * What sequences change between cgram, ddram access?


 * Read behavior: After a set address instruction content from
 * ram is moved to DR, then the BF is deasserted. The next read
 * will return the data from DR immediatly, assert the BF and
 * fetch the next data if increment mode active. After BF gets
 * low, another 1.5 cycles are needed until AC is updated.
 *
 * getNextStep stores the next state in variables, if a doSteps
 * with same length follows the states calculated in getNextStep
 * get written into the current variables
 */



public class LCDDisplay2 implements fips.sim.dig.base.Component {

  String				name;

  public final static int		PIN_DATA0	=  0;
  public final static int		PIN_DATA7	=  7;
  public final static int		PIN_ENABLE	=  8;
  public final static int		PIN_RW		=  9;
  public final static int		PIN_RS		= 10;
  public final static int		PIN_CONTRAST	= 11;
  public final static int		PIN_HILIGHT	= 12;

  fips.sim.dig.comp.Pin			pins[];
  int					pinnums[];
  int					states[];

/** Requests for pin changes, -1 if unchanged
 */
  int					nextstates[];

/** List of pins that have to remain unchanged. Changes on these pins
 *  will result in signal violation errors.
 */
  boolean				holds[];

  boolean				nextholds[];


/** Values of the sampled pins.
 */
  int					holdstates[]; 


/** State of each pin before current state was reached
 */
  int					laststates[];

/** Time elapsed since pin's state change from laststate to current
    state. This value can't get bigger than maxvalue
 */
  long					lastchgs[];



/**Scalefactor to allow bigger display
*/
  int					scalef;

  LCDDisplayGUI				lcdgui;



// Component specific variables


/** Current cursor pos in ddram
*/
  int					ramaddr;


/** If true a read or write command increments ram address
 */
  boolean				incaddrmode;


/** If true ram rws affect the display data ram.
 */
  boolean				ddrammode;
/** If true ram rws affect the character generator ram.
 */
  boolean				cgrammode;


/** State of the busy flag in the controller. After a read or
 *  write ram instruction, the flag may be off but the busytimer
 *  set while controller changes ddram address value
 */
//  boolean				busyflag;


/** State of the last command executed. This is needed to allow
 *  additional busy cycles with busyflag false after an instruction
 *  increasing the address register.
 */
  int					lastaction;


/** State of the function set lines at the begin of an instruction.
 *  The state is reduced to -1, 0, 1 for unknown, low and high.
 */
  int					rwline;
  int					rsline;



/** If true any exceeding of maximum/minimum timing specifications
 *  will result in a warning message. Also an unknown signal
 *  that is longer than a spike will result in a message.
 */
  boolean				strict;




// uninitialized, EN low
  final static int			STATE_UNINIT	= 0;

/** Entry point for normal device reset, just puts busy flag to
 *  high. */
  final static int			STATE_RESET	= 0x100;

/** Wait reset time. */
  final static int			STATE_RESETWAIT	= 0x101;


// en low, doing nothing
  final static int			STATE_IDLE	= 0x110;

  final static int			STATE_READRAM	= 0x200;
  final static int			STATE_WRITERAM	= 0x210;

  final static int			STATE_MASK	= 0xFFF0;



/** Current controller state of the lcd panel
 */
  int					lcdstate;


  LCDControler				control; 


// timing critical values


  long					nextstep;


/** Absolute running time of this device
 */
  java.math.BigInteger			time;


/** Length of the current state or -1 if indefinitly long.
 */
  long					statelength;




/** Maximum length of an unknown signal on a pin until a state
    may get triggered. The default value is 1 ns. 
 */
  long					spike=1000000;


/** Internal clock period, currently 4us (250kHz)
 */
  long					intclock=4000000000L;


/** Time for a reset performance, max 10ms
 */
  long					resettime=10000000000000L;

/** If true, the unit performs a normal reset, no init by
    instruction needed
 */
  boolean				normalreset;





/** Minimum valid cycle time from one to next rise of enable
    pin: 500ns at 4.5V, 1000ns at 2.3V
 */
  long					tcycmin=500000000L;

  java.math.BigInteger			cycstart;


/** Max time for an enable rise or fall, 20ns at 4.5V, 25ns
    at 2.3 V
 */
  long					enableedgemax=20000000L;


/** Cycles since start of enable edge
 */
  long					enableedgetime;


/** Minimum pulse width of enable signal: 230ns at 4.5V, 450 at 2.3V
 */
  long					tenablepw=230000000L;

/** Time of the last enable pin transition to valid high or
    low state.
 */
  long					enplength;



/** Time before rise of enable that the address lines have
 *  to be valid: 40ns at 4.5V, 60ns at 2.3V
 */
  long					taddrset=40000000L;


/** Time after falling edge of en the address lines have to stay
 *  valid: 10ns at 4.5V, 20ns at 2.3V
 */
  long					taddrhold=10000000L;


/** Time from begin of read instruction (en high) until bus data
 *  is valid: 160ns at 4.5V, 360 at 2.3V
 */
  long					tdataready=160000000L;                          


/** Time read data stays ready after en got low again:
 *  5ns at 4.5V, 10ns at 2.3V
 */
  long					tdatahold=5000000L;






  public LCDDisplay2() {
    int i;

    name=getComponentName();

// initialize the pins
    pins=new fips.sim.dig.comp.Pin[13];
    pinnums=new int[13];
    states=new int[13];
    nextstates=new int[13];
    laststates=new int[13];
    lastchgs=new long[13];
    holds=new boolean[13];
    nextholds=new boolean[13];

    for(i=0; i<states.length; i++) {
      states[i]=Pin.HiZ;
      laststates[i]=Pin.UNKNOWN;
      lastchgs[i]=0;
      holds[i]=nextholds[i]=false;
      nextstates[i]=-1;
    }

    lcdgui=new LCDDisplayGUI();

// set preferred scale, scalefactor and real diplay type
    lcdgui.setType(4, 1, lcdgui.TYPE_SMALL2x16);


    normalreset=true;
    if(normalreset) {
      lcdstate=STATE_RESET;
// set the busy flag
    } else {
      lcdstate=STATE_UNINIT;
    }

    statelength=0;

    time=java.math.BigInteger.ZERO;
    control=new LCDControler(lcdgui);

// increment ramaddr ?
    incaddrmode=true;

    cgrammode=false;
    ddrammode=false;


    cycstart=time;

    lastaction=0;

    nextstep=-1;
  }




  public String getComponentName() {
    return("LCDDisplay");
  }




  public String getDescription() {
    return("LCD Display");
  }


  public String getPinDescription(int num) {
    throw new InternalError();
  }

  public String getName() {
    return(name);
  }
 
 
  public void setName(String name) {
    this.name=name;
  }

  public int getParameterCount() {
    throw new InternalError();
  }

  public fips.sim.dig.base.Parameter[] getParameters() {
    throw new InternalError();
  }


  public void setParameter(int num, Object val) {
    throw new InternalError();
  }



  public int getPinCount() {
    return(13);
  }




  public void setPin(int num, fips.sim.dig.comp.Pin p) {
    if(num>=13) throw new IllegalArgumentException("Can't set pin "+num);

    if(pins[num]!=null) pins[num].removeComponent(pinnums[num]);

    pins[num]=p;
    if(pins[num]!=null) {
      pinnums[num]=pins[num].addComponent();
      pins[num].setValue(Pin.HiZ, pinnums[num]);
    }
  }






  public fips.sim.dig.comp.Pin getPin(int num) {
    if(num>=13) throw new IllegalArgumentException();
    return(pins[num]);
  }









/** Get a hint for max length of next step. This method just
    calculates the step length and sets variables due to pin
    changes, but won't alter any state or pin itself.
 */
  public long getNextStep() {
    int				i, j;
    long			l;
    java.math.BigInteger	bi;

    nextstep=Long.MAX_VALUE;

    for(i=0; i<pins.length; i++) {
      j=pins[i].getValue();
      if(j!=states[i]) {
        if(getReducedPinState(j)==getReducedPinState(states[i])) {
// no real change
          states[i]=j;
        } else {
          laststates[i]=states[i];
          states[i]=j;
          lastchgs[i]=0;
        }
      }
      if(isPinUnknown(states[i])&&(lastchgs[i]<spike)) {
        l=spike-lastchgs[i];
        if(l<nextstep) nextstep=l;
      }
    }


    control.setValues(states);

// check the hold pins for continuity
// changes shorter than spike are ignored
    for(i=0; i<holds.length; i++) {
      if((!holds[i])||(states[i]==holdstates[i])) continue;
      if(getReducedPinState(states[i])==getReducedPinState(holdstates[i])) {
        holdstates[i]=states[i];
        continue;
      }

      if(isPinUnknown(states[i])&&(lastchgs[i]<spike)) continue;

      println("Illegal state change for pin "+i+" from "+holdstates[i]+" to "+
              states[i]);
      holdstates[i]=states[i];
    }





    l=control.getNextStep();
    if(l<nextstep) nextstep=l;


    switch(lcdstate) {
      case STATE_RESET:
        nextstates[PIN_DATA7]=Pin.HIGH;
        statelength=0;
        break;
      case STATE_RESETWAIT:
        if(statelength==0) {
          statelength=resettime;
          nextstates[PIN_DATA7]=Pin.LOW;
        }
        break;
      case STATE_IDLE:
        if((statelength==0)||(lastchgs[PIN_ENABLE]==0)) { // pin changed
          if(isPinUnknown(states[PIN_ENABLE])) {
            statelength=spike;
            cycstart=time;
          } else if(isPinHigh(states[PIN_ENABLE])) {
            statelength=0;
            cycstart=time;
          } else {	// pin low
            statelength=Long.MAX_VALUE;
          }
        }
        //ARM ("#: leaving idle state in "+statelength+"fs, current pin state "+
        //ARM  states[PIN_ENABLE]+".");
        break;
      default:
        throw new InternalError("Invalid state 0x"+
                                 Integer.toHexString(lcdstate));
    }

    if((statelength>=0)&&(statelength<nextstep)) nextstep=statelength;

    return(nextstep);
  }









/** Do the following calculation step. If cycles is too big,
    IllegalArgumentException is thrown
 */
  public void doStep(long cycles) {
    int				i, n;
    java.math.BigInteger	bi;


    for(i=0; i<lastchgs.length; i++) {
      lastchgs[i]+=cycles;
      if(lastchgs[i]<0) lastchgs[i]=Long.MAX_VALUE;	// overflow
    }

    if(cycles>0) time=time.add(java.math.BigInteger.valueOf(cycles));

    control.doStep(cycles);

    if(statelength>=0) {
      if(cycles>statelength) throw new InternalError();
      statelength-=cycles;
    }

    if(nextstep<0) throw new InternalError();

    if(cycles>nextstep) throw new InternalError();
    nextstep-=cycles;

    if(statelength==0) {
// copy newvars to vars

      for(i=0; i<nextstates.length; i++) {
        if(nextstates[i]==-1) continue;

        pins[i].setValue(nextstates[i], pinnums[i]);
        nextstates[i]=-1;
      }

      for(i=0; i<nextholds.length; i++) {
        if(holds[i]==nextholds[i]) continue;

        if(holds[i]==true) {
          //ARM ("#: sample/hold on pin "+i+" deactivated.");
          holds[i]=false;
          continue;
        }
        holds[i]=true;
        //ARM ("#: sample/hold on pin "+i+" activated.");
        holdstates[i]=states[i];
        if(isPinUnknown(states[i])) {
          //ERROR ("Sampled unknown value on pin "+i+", assuming 0");
          holdstates[i]=0;
        }
      }


// enter next state
      nextstep=-1;

      switch(lcdstate) {
        case STATE_RESETWAIT:
          //ARM ("#: reset complete");
          lcdstate=STATE_IDLE;
          break;
        default:
          //ARM ("#: no special state handling needed, going to next state");
          lcdstate++;
      }

    }
  }









/**Reset cyclecount, params, etc...
*/
  public void reset() {
    throw new InternalError();
  }




  public void writeState(java.io.ObjectOutputStream oos)
                         throws java.io.IOException {
    throw new InternalError();
  }


  public void readState(java.io.ObjectInputStream ois)
                        throws java.io.IOException {
    throw new InternalError();
  }




/**Get the GUI size of this component.
*/
  public java.awt.Dimension getSize() {
    return(lcdgui.getSize());
  }


  public void setGUIParameters(java.awt.Point offs, int scale, int rot) {
    throw new InternalError();
  }


  public int getOrientation() {
    throw new InternalError();
  }


/** Init the graphics with given scale and calculate the values
 *  of boarders and spaces. 
 */
  public void repaint(Graphics g) {
  }

  public void update(Graphics g) {
  }

  public boolean needsUpdate() { return(true); }



/** Take actions if clicked, currently none
*/
  public void doClick(int x, int y, int button) {
  }



  public java.awt.Point getPinPos(int num) {
    throw new InternalError();
  }


  public Object clone() {
    throw new InternalError();
  }




// ************************************************************************* //
//
//			Internal Methods				     //
//
// ************************************************************************* //


  void println(String text) {
    System.err.println(name+": "+text);
  }




// ************************************************************************* //
//
//			Pin state changes				     //
//
// ************************************************************************* //



  boolean isPinHigh(int var) {
    return((var==Pin.HIGH)||(var==Pin.WEAKHIGH));
  }

  boolean isPinLow(int var) {
    return((var==Pin.LOW)||(var==Pin.WEAKLOW));
  }

  boolean isPinUnknown(int val) {
    return((val==Pin.UNKNOWN)||(val==Pin.WEAKUNKN)||(val==Pin.HiZ));
  }


/** Return the simplified pin state for a given pin. If the
 *  pin value is driven hard or weak, the state is returned, if
 *  undecidable the result is -1
 */
  int getReducedPinState(int val) {
    if((val==Pin.HIGH)||(val==Pin.WEAKHIGH)) return(1);
    if((val==Pin.LOW)||(val==Pin.WEAKLOW)) return(0);
    return(-1);
  }


}






