package fips.sim.dig.comp;

/** Class for connecting numerous pins together. If the
 *  requestcount is 0, the pin waits for all attached
 *  components to make their requestStep calls, if
 *  requcnt==# of components the pin transmitts the
 *  current value to all components.
 */

public class Pin {

  public final static int		LOW	= 0;
  public final static int		HIGH	= 1;
  public final static int		UNKNOWN	= 2;	// collision
  public final static int		HiZ	= 3;
  public final static int		WEAKLOW	= 4;
  public final static int		WEAKHIGH = 5;
  public final static int		WEAKUNKN = 6;


  final static int transitions[][] = {
    {LOW,     UNKNOWN, UNKNOWN, LOW,      LOW,      LOW,      LOW },
    {UNKNOWN, HIGH,    UNKNOWN, HIGH,     HIGH,     HIGH,     HIGH},
    {UNKNOWN, UNKNOWN, UNKNOWN, UNKNOWN,  UNKNOWN,  UNKNOWN,  UNKNOWN},
    {LOW,     HIGH,    UNKNOWN, HiZ,      WEAKLOW,  WEAKHIGH, WEAKUNKN},
    {LOW,     HIGH,    UNKNOWN, WEAKLOW,  WEAKLOW,  WEAKUNKN, WEAKUNKN},
    {LOW,     HIGH,    UNKNOWN, WEAKHIGH, WEAKUNKN, WEAKHIGH, WEAKUNKN},
    {LOW,     HIGH,    UNKNOWN, WEAKUNKN, WEAKUNKN, WEAKUNKN, WEAKUNKN}
  };


/** Number of the highest assigned state variable.
 */
  int				maxstate;
  int				states[];
  boolean			changed;
  int				state;

  public Pin() {
    state=HiZ;
    maxstate=0;
    states=new int[8];
    changed=false;
  }


  public synchronized int getComponentCount() {
    int cnt=maxstate;
    for(int i=0; i<maxstate; i++) {
      if(states[i]<0) cnt--;
    }
    return(cnt);
  }



/** Request access to this pin connector and get registration number
 */
  public synchronized int addComponent() {

    if(maxstate<states.length) {
      int i=maxstate;
      states[i]=HiZ;
      maxstate++;
      changed=true;
      return(i);
    }

    for(int i=0; i<maxstate; i++) {
      if(states[i]<0) {
        states[i]=HiZ;
        changed=true;
        return(i);
      }
    }

// enlarge arrays
    throw new InternalError("No array enlarge yet");
  }



  public synchronized void removeComponent(int num) {
    if(states[num]<0) throw new IllegalArgumentException();

    states[num]=-1;
    changed=true;

    while((maxstate>0)&&(states[maxstate-1]==-1)) maxstate--;
    return;
  }



/** Get the current value of this pin
 */
  public synchronized int getValue() {
    if(changed) {
      state=HiZ;
      for(int i=0; i<maxstate; i++) {
        if(states[i]<0) continue;
        state=transitions[state][states[i]];
      }

      changed=false;
    }

    return(state);
  }


/** Set value for given position.
 */
  public synchronized void setValue(int val, int num) {
    if((states[num]<0)||(val<0)||(val>WEAKUNKN)) {
      throw new IllegalArgumentException("Illegal setValue(val="+val+", num="+
                                         num+").");
    }

    if(states[num]!=val) {
      states[num]=val;
      changed=true;
    }
  }



  private final static int redstates[]={0, 1, -1, -1, 0, 1, -1};
  public static int getReducedState(int val) {
    return(redstates[val]);
  }
}



