package fips.sim.dig.comp.fsm;


public class FSMachine  {

  protected int		statevars[];
  protected int		newvars[];
  long			timers[];
  long			newtimers[];

  java.util.Vector	groups;


/** In the time from entering getNextStep to leaving doSteps
 *  holdvars will be true to avoid modifications of the current
 *  state variables which would lead to unpredictable behaviour.
 */
  boolean		holdvars;


  public FSMachine(int vars, int timers) {
    statevars=new int[vars];
    newvars=new int[vars];
    this.timers=new long[timers];
    newtimers=new long[timers];

    groups=new java.util.Vector();

    for(int i=0; i<vars; i++) newvars[i]=fips.sim.dig.comp.Pin.UNKNOWN;

    holdvars=false;
  }



/** Create a new state group with a given name.
 */
  public void createStateGroup(String name) {
    //TRACE (".createStateGroup("+name+").");
    if(groups.contains(name)) {
      throw new IllegalArgumentException("FSMachine allready contains stategroup with name "+name);
    }

    groups.addElement(new StateGroup(name));
  }





  public void addState(StateDef s, String group) {
    StateGroup		g;

    if((g=getGroup(group)).states.get(s.getName())!=null) {
      throw new IllegalArgumentException("Can't register two states with name "+
                                         s.getName());
    }

    g.states.put(s.getName(), s);
  }
  


/** Change the initial state in one stategroup. Setting the
 *  the state is the same as an unconditional state transition
 *  from the current to the named state, the machine will
 *  leave the current state with current.leave() and enter the
 *  new one with enter(). To work correctly all variables needed
 *  by enter() or leave() have to be set prior to calling this
 *  method.
 */
  public void setInitialState(String name, String group) {
    StateGroup	g;

    g=(StateGroup)getGroup(group);
    g.current=(StateDef)g.states.get(name);
    if(g.current==null) {
      throw new IllegalArgumentException("No state "+name+" in group "+group);
    }
  }



  public StateDef getStateForName(String name, String group) {
    return((StateDef)getGroup(group).states.get(name));
  }




/** Method to set variables from outside the final state machine.
 *  The value is written into the newvars and currentvars arrays
 *  after leaving doSteps and before entering getNextStep, in
 *  the remaining cycle the value is only written into the
 *  newvars and will get copied to the current vars in the next
 *  getNextStep call.
 */
  public synchronized void setValue(int val, int pos) {
    if(!holdvars) statevars[pos]=val;
    newvars[pos]=val;
  }




  public synchronized long getNextStep() {
    long		l, m;
    int			i;
    Transition		t;

    holdvars=true;

// move in last doStep() call changed variable values to statevars
    System.arraycopy(newvars, 0, statevars, 0, statevars.length);
    System.arraycopy(newtimers, 0, timers, 0, timers.length);

    l=Long.MAX_VALUE;
    for(i=0; i<groups.size(); i++) {

      t=((StateGroup)groups.elementAt(i)).current.getNextTransition();

      if(t==null) continue;

      if((m=t.getTime())<l) l=m;
      //ARM ("#: max step for group "+i+" is "+m);
    }
    return(l);
  }





  public synchronized void doStep(long cycles) {
    int				i;

    //TRACE (".doStep("+cycles+").");

    if(!holdvars) throw new InternalError("Synchronization error");

    for(i=0; i<groups.size(); i++) {
      ((StateGroup)groups.elementAt(i)).current.doSteps(cycles);
    }
    holdvars=false;
  }






  private StateGroup getGroup(String group) {
    StateGroup		g;
    int			i;

    for(i=0; i<groups.size(); i++) {
      if((g=(StateGroup)groups.elementAt(i)).name.equals(group)) return(g);
    }

    throw new IllegalArgumentException("No stategroup named "+group);
  }





  public void doTransition(Transition t, StateDef newstate) {
    StateGroup	g;
    int		i;

    g=getGroup(newstate.getGroup());
    if(t.changeactions!=null) {
      for(i=0; i<t.changeactions.length; i++) {
        t.changeactions[i].performAction(this, 0);
      }
    }
    if(t.changemethod!=null) {
      t.changemethod.invoke();
    }
    g.current=newstate;
    //TRACE ("#: entered state "+newstate.getName()+"("+
    //TRACE  newstate.getGroup()+").");
  }


}




class StateGroup {
  String			name;
  java.util.Hashtable		states;
  StateDef			current;

  public StateGroup(String name) {
    this.name=name;
    states=new java.util.Hashtable();
    current=null;
  }


  public String toString() { return(name); }


}







