package fips.sim.dig.comp.fsm;


/** Definition of a state for a final state machine. Entering,
 *  leaving and moving in the state may generate method calls
 *  or variable changes. The state can be left when one or more
 *  ChangeConditions get true. The state may not contain an
 *  explicit definition of a loopback ChangeCondition, it would
 *  spoil the timing generation functionality. To stay in a state
 *  if no ChangeCondition matches is defined as implicit behavior.
 */
public class StateDef {

  private String			name;
  private String			group;

  private FSMachine			fsmachine;

  public StateAction			stayactions[];
  public ActionMethod			staymethod;


/** State transitions defined for this state. The array may be
 *  longer than the actual number of transitions, a null value
 *  marks the last transition element.
 */
  private Transition			transitions[];
  private StateDef			targets[];


/** Transitions to this state
 */
  private Transition			reftransitions[];
  private StateDef			sources[];


  public StateDef(FSMachine fsm, String name, String group) {
    this.name=name;
    this.fsmachine=fsm;
    this.group=group;

    fsm.addState(this, group);
  }


  public String getName() { return(name); }

  public String getGroup() { return(group); }


/** Do steps and state transitions.
 */
  public void doSteps(long cycles) {
    int		i;

    if((cycles!=0)&&(stayactions!=null)) {
      for(i=0; i<stayactions.length; i++) {
        stayactions[i].performAction(fsmachine, cycles);
      }
    }

    if(staymethod!=null) {
      throw new InternalError();
    }


    for(i=0; i<transitions.length; i++) {
      if(transitions[i]==null) break;
      if(transitions[i].getTime()==0) {
        fsmachine.doTransition(transitions[i], targets[i]);
      }
    }
  }




/** Get the next Transition from this state if the values stay
 *  the same.
 */
  public Transition getNextTransition() {
    int		i, next, staycond;
    long	l, steps;

    next=-1;
    steps=Long.MAX_VALUE;

    for(i=0; i<transitions.length; i++) {
      //ARM ("#: getting time for transition "+i);
      if(transitions[i]==null) break;

      l=transitions[i].getTime();
      if(l==-1) continue;			// can't get true

      if(l<0) throw new InternalError();	// not val counters

      if(l<steps) {
        steps=l;
        next=i;
      } else if((l==steps)&&(steps!=Long.MAX_VALUE)) {
        throw new InternalError();
      }
    }

    return( next<0 ? null: transitions[next] );
  }



/** Add a transition to this state. This also adds the transition to
 *  the reftransitions of the nextstate
 */
  public void addTransition(Transition t, StateDef nextstate) {
    int pos;

    pos=0;
    if(transitions==null) {
      transitions=new Transition[8];
      targets=new StateDef[8];
      transitions[0]=t;
      targets[0]=nextstate;
      
    } else {
      if(reftransitions[reftransitions.length-1]!=null) {
        throw new InternalError();
      }

      for(pos=0; pos<transitions.length; pos++) {
        if(transitions[pos]==null) break;

        if((transitions[pos].equals(t))&&(targets[pos].equals(nextstate))) {
          throw new IllegalArgumentException("Multiple state links");
        }
      }
    }

    nextstate.addRefTransition(t, this);
    transitions[pos]=t;
    targets[pos]=nextstate;
  }



  final private void addRefTransition(Transition t, StateDef source) {
    int		i;

    if(reftransitions==null) {
      reftransitions=new Transition[8];
      sources=new StateDef[8];
      reftransitions[0]=t;
      sources[0]=source;
      return;
    }

    if(reftransitions[reftransitions.length-1]!=null) {
// enlarge array
      throw new InternalError();
    }

    for(i=0; i<reftransitions.length; i++) {
      if(reftransitions[i]==null) break;
    }
    reftransitions[i]=t;
    sources[i]=source;
  }


}



