package fips.sim.dig.edit;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;

import fips.sim.dig.edit.rcfile.DigSimConfig;
import fips.sim.dig.edit.rcfile.GateGroupType;
import fips.sim.dig.edit.rcfile.GateType;
import fips.sim.dig.edit.rcfile.ObjectFactory;
import fips.sim.dig.edit.rcfile.ParameterGroupType;
import fips.sim.dig.edit.rcfile.ParameterType;



/** The parameters are kept unsorted, the gates are treated
 *  like a arraylist. When merging parent gate-trees with
 *  new, structure of new is filled with old gates (if possible)
 */

public class RCFile extends Thread {

  Logger		logger;

  String		urlstring;

  RCFile		parentrc;


/** For autosync error messages.
 */
  StringBuffer		errors;

/** True if changes should be written to this file
 */
  boolean		writemode;

/** If true a setparam will cause the rcfile to write if any
 *  parameter changes.
 */
  boolean		autowrite;

  boolean		changed;


/** The toplevel root group holding all parameters and
 *  and other subgroups.
 */
  DigSimConfig		content;



/** Create an empty rcfile. This is usefull to let setparam
 *  calls even in a readonly environment work. It may also be
 *  used to collect usersettings and assign a destinationurl
 *  later (not yet impl).
 */
  public RCFile(boolean writemode, String urlstr, RCFile parentrc)
                throws IllegalArgumentException, IOException {
    logger=Logger.getLogger("digsim.edit");
    logger.fine(".init(writemode="+writemode+", urlstr="+urlstr+", parentrc="+
                parentrc+").");

    this.writemode=writemode;
    autowrite=true;
    changed=false;

    this.parentrc=parentrc;

    if(urlstr!=null) {
      if(urlstr.startsWith("sys://")) {
        if(writemode) throw new IllegalArgumentException("Writemode for system resources not allowed.");
      } else if(urlstr.startsWith("file://")) {
        logger.severe("No modecheck yet..");
      } else {
        throw new IllegalArgumentException("Illegal url type in url "+urlstr);
      }
      loadUrl(urlstr);
      start();
    }
    this.urlstring=urlstr;

  }






//**************************************************************************//
//									    //
//				Data Handling				    //
//									    //
//**************************************************************************//


  public ParameterGroup getParameterGroup(String path[]) {
    if((path==null)||(path.length==0)) return(new ParameterGroup(this, null));

    if(content!=null) { // search for group in content
      Object current=content;
      ParameterGroupType pg=null;

      for(int pathcnt=0; pathcnt<path.length; pathcnt++) {
        pg=getGroup(current, path[pathcnt]);
        if(pg==null) break;
        current=pg;

        if(pathcnt==path.length-1) {
// end of path reached, group found -> return object
          return new ParameterGroup(this, path);
        } 
      }

      if(content.isIgnoreParentParams()) return(null);
    }

    if((parentrc!=null)&&(parentrc.getParameterGroup(path)!=null)) {
      return(new ParameterGroup(this, path));
    }

    return(null);
  }


/** Create a new parametergroup tree.
 */
  public void createParameterGroup(String path[]) {

    if((path==null)||(path.length==0)) {
      throw new IllegalArgumentException("Can't create root groups.");
    }

    if(content==null) initContent();

    Object current=content;
    for(int pathcnt=0; pathcnt<path.length; pathcnt++) {
      ParameterGroupType pg=createGroup(current, path[pathcnt]);
      current=pg;
    }

    changed=true;
  }



  public String getParameter(String path[], String param) {
    if(content!=null) {
      Object current=content;

      for(int pathcnt=0; pathcnt<path.length; pathcnt++) {
        current=getGroup(current, path[pathcnt]);
        if(current==null) break;

        if(pathcnt+1==path.length) {
          ParameterType pt=getParameter(current, param);
          if(pt!=null) return(pt.getValue());
        }
      }

// not found
      if(content.isIgnoreParentParams()) return(null);
    }

    if(parentrc!=null) return(parentrc.getParameter(path, param));

    return(null);
  }



  public void setParameter(String path[], String param, String value) {
    logger.fine(".setParameter(path, param="+param+", value="+value+").");

    if(!writemode) throw new IllegalArgumentException();
    if(content==null) initContent();

    Object current=content;

    for(int pathcnt=0; pathcnt<(path==null?0:path.length); pathcnt++) {
      current=createGroup(current, path[pathcnt]);
    }

    ParameterType pt=createParameter(current, param);
    pt.setValue(value);

    changed=true;
  }









  private ParameterGroupType getGroup(Object parent, String key) {
    List list = null;

    if(parent instanceof DigSimConfig) {
      list=((DigSimConfig)content).getGateOrGateGroupOrParameter();
    } else if(parent instanceof ParameterGroupType) {
      list=((ParameterGroupType)parent).getParameterGroupOrParameter();
    } else throw new IllegalArgumentException();

    for(int i=0; i<list.size(); i++) {
      if(!(list.get(i) instanceof ParameterGroupType)) continue;

      if(((ParameterGroupType)list.get(i)).getKey().equals(key)) {
        return((ParameterGroupType)list.get(i));
      }
    }
    return(null);
  }



  private ParameterType getParameter(Object parent, String param) {
    List list = null;

    if(parent instanceof DigSimConfig) {
      list=((DigSimConfig)content).getGateOrGateGroupOrParameter();
    } else if(parent instanceof ParameterGroupType) {
      list=((ParameterGroupType)parent).getParameterGroupOrParameter();
    } else throw new IllegalArgumentException();

    for(int i=0; i<list.size(); i++) {
      if(!(list.get(i) instanceof ParameterType)) continue;

      if(((ParameterType)list.get(i)).getName().equals(param)) {
        return((ParameterType)list.get(i));
      }
    }
    return(null);
  }


/** Locate or create the gategroup
 */
  private ParameterGroupType createGroup(Object parent, String key) {
    if(!writemode) throw new IllegalArgumentException();

    List list = null;

    if(parent instanceof DigSimConfig) {
      list=((DigSimConfig)content).getGateOrGateGroupOrParameter();
    } else if(parent instanceof ParameterGroupType) {
      list=((ParameterGroupType)parent).getParameterGroupOrParameter();
    } else throw new IllegalArgumentException();

    for(int i=0; i<list.size(); i++) {
      if(!(list.get(i) instanceof ParameterGroupType)) continue;

      if(((ParameterGroupType)list.get(i)).getKey().equals(key)) {
        return((ParameterGroupType)list.get(i));
      }
    }


    try {
      ParameterGroupType pg = ObjectFactory.createParameterGroupType();

      pg.setKey(key);
      list.add(pg);
/*
      if(parent instanceof DigSimConfig) {
        ((DigSimConfig)parent).setGateOrGateGroupOrParameter(list);
      } else {
        ((ParameterGroupType)parent).setParameterGroupOrParameter(list);
      }
*/
      return(pg);
    } catch (JAXBException jbe) {
      //ERROR ("Failed to create new DigSimConfig object.");
      jbe.printStackTrace();
      throw new InternalError();
    }
  }





  private ParameterType createParameter(Object parent, String param) {
    if(!writemode) throw new IllegalArgumentException();

    List list = null;

    if(parent instanceof DigSimConfig) {
      list=((DigSimConfig)content).getGateOrGateGroupOrParameter();
    } else if(parent instanceof ParameterGroupType) {
      list=((ParameterGroupType)parent).getParameterGroupOrParameter();
    } else throw new IllegalArgumentException();

    for(int i=0; i<list.size(); i++) {
      if(!(list.get(i) instanceof ParameterType)) continue;

      if(((ParameterType)list.get(i)).getName().equals(param)) {
        return((ParameterType)list.get(i));
      }
    }


    try {
      ParameterType pt = ObjectFactory.createParameterType();

      pt.setName(param);
      list.add(pt);
/*
      if(parent instanceof DigSimConfig) {
        ((DigSimConfig)parent).setGateOrGateGroupOrParameter(list);
      } else {
        ((ParameterGroupType)parent).setParameterGroupOrParameter(list);
      }
*/
      return(pt);
    } catch (JAXBException jbe) {
      //ERROR ("Failed to create new DigSimConfig object.");
      jbe.printStackTrace();
      throw new InternalError();
    }

  }




//**************************************************************************//
//
//			GateGroup handling
//
//**************************************************************************//





  public GateGroup getGateGroup(String path[]) {
    if((path==null)||(path.length==0)) return(new GateGroup(this, null, null));

    if(content!=null) { // search for group in content
      Object current=content;
      for(int pathcnt=0; pathcnt<path.length; pathcnt++) {
        current=getGateGroup(current, path[pathcnt]);
        if(current==null) break;
      }

      if(current!=null) {
        return(new GateGroup(this, path, ((GateGroup)current).getName()));
      }

      if(content.isIgnoreParentGates()) return(null);
    }

    if(parentrc!=null) {
      GateGroup gg=parentrc.getGateGroup(path);
      if(gg!=null) return(new GateGroup(this, path, gg.getName()));
    }

    return(null);
  }


/** 
 *  @returns null if the path was not found
 */
  public Object[] getGateGroupContent(String path[]) {
    logger.fine(".getGateGroupConent(path)");

    List list=null;
    Object pdata[]=null;
    Object current=content;

    if(content!=null) {

      for(int pathcnt=0; pathcnt<(path==null?0:path.length); pathcnt++) {
        current=getGateGroup(current, path[pathcnt]);
        if(current==null) break;
      }
    }

    logger.finest("Content ignoreparentgates="+content.isIgnoreParentGates()+
                  ", parent "+parentrc);
    if(((content==null)||(!content.isIgnoreParentGates()))&&(parentrc!=null)) {
      logger.finer("Consulting parent rcfile for gate-content.");
      pdata=parentrc.getGateGroupContent(path);
    }

    if(current!=null) {
      if(current instanceof DigSimConfig) {
        list=((DigSimConfig)current).getGateOrGateGroupOrParameter();
      } else list=((GateGroupType)current).getGateGroupOrGate();
    }

    logger.fine("Using "+(list!=null?list.size():-1)+" elems from current rc, "+
                (pdata==null?-1:pdata.length)+" elements from parent.");

    if(list==null) return(pdata);

    ArrayList ret=new ArrayList();

    for(int i=0; i<list.size(); i++) {
      if((list.get(i) instanceof fips.sim.dig.edit.rcfile.GateGroupType)) {
        fips.sim.dig.edit.rcfile.GateGroupType gt=
           (fips.sim.dig.edit.rcfile.GateGroupType)list.get(i);
        logger.finer("Adding new gategroup "+gt.getKey());
 
        String np[]=new String[(path==null?1:path.length+1)];
        if(path!=null) System.arraycopy(path, 0, np, 0, path.length);
        np[np.length-1]=gt.getKey();

        GateGroup gg=new GateGroup(this, np, gt.getName());
        ret.add(gg);
        if(pdata!=null) {
          for(int j=0; j<pdata.length; j++) {
            if((pdata[j] instanceof GateGroup)&&
               (((GateGroup)pdata[j]).getKey().equals(gt.getKey()))) {
              pdata[j]=null;
              break;
            }
          }
        }
      }
      if((list.get(i) instanceof fips.sim.dig.edit.rcfile.GateType)) {
        fips.sim.dig.edit.rcfile.GateType gt=
              (fips.sim.dig.edit.rcfile.GateType)list.get(i);
        ret.add(new Gate(gt.getName(), gt.getKey(), gt.getClassName()));

        if(pdata!=null) {
          for(int j=0; j<pdata.length; j++) {
            if((pdata[j] instanceof Gate)&&
               (((Gate)pdata[j]).key.equals(gt.getKey()))) {
              pdata[j]=null;
              break;
            }
          }
        }
      }
    }

    if(pdata!=null) {
      for(int i=pdata.length-1; i>=0; i--) {
        if(pdata[i]!=null) ret.add(0, pdata[i]);
      }
    }

    pdata=new Object[ret.size()];
    ret.toArray(pdata);

    logger.finer("Content of group "+path+" is "+pdata.length);

    return(pdata);
  }








  private GateGroupType getGateGroup(Object parent, String key) {
    List list = null;

    if(parent instanceof DigSimConfig) {
      list=((DigSimConfig)content).getGateOrGateGroupOrParameter();
    } else if(parent instanceof GateGroupType) {
      list=((GateGroupType)parent).getGateGroupOrGate();
    } else throw new IllegalArgumentException();

    for(int i=0; i<list.size(); i++) {
      if(!(list.get(i) instanceof GateGroupType)) continue;

      if(((GateGroupType)list.get(i)).getKey().equals(key)) {
        return((GateGroupType)list.get(i));
      }
    }
    return(null);
  }





// If content is null create an empty content
  private void initContent() {
    if(content!=null) return;

    try {
      content = fips.sim.dig.edit.rcfile.ObjectFactory.createDigSimConfig();
    } catch (JAXBException jbe) {
      //ERROR ("Failed to create new DigSimConfig object.");
      jbe.printStackTrace();
      throw new InternalError();
    }
  }



//**************************************************************************//
//
//			AutoRead/Write Features				    //
//
//**************************************************************************//



  public String getErrors() {
    if((errors!=null)&&(errors.length()!=0)) {
      String e=errors.toString();
      errors.setLength(0);
      return(e);
    }
    return(null);
  }



  public void setAutoWrite(boolean val) {
    autowrite=val;
  }






//**************************************************************************//
//									    //
//			Private members					    //
//									    //
//**************************************************************************//


/** If url is fileurl and file doesn't exist -> load nothing
 */
  private synchronized void loadUrl(String url) throws java.io.IOException,
                       IllegalArgumentException {

    InputStream input=null;
    content=null;

    if(url.startsWith("sys://")) {

      input=getClass().getResourceAsStream(url.substring(6));
      if(input==null) {
        input=ClassLoader.getSystemResourceAsStream(url.substring(6));
      }

      if(input==null) {
        //ERROR ("Failed to open systemresource "+url.substring(6));
        return;
      }

    } else if(url.startsWith("file://")) {
      if(!new java.io.File(url.substring(7)).exists()) return;
      input = new java.io.FileInputStream(url.substring(7));
    } else {
      throw new IllegalArgumentException("Illegal url "+url);
    }

// now read the inputstream data
    try {
      JAXBContext jc = JAXBContext.newInstance("fips.sim.dig.edit.rcfile");

      Unmarshaller u = jc.createUnmarshaller();

      content = (DigSimConfig)u.unmarshal(input);
    } catch (JAXBException jbe) {
      jbe.printStackTrace();
      throw new IOException("Unmarshalling of xml-data failed");
    }
  }






  private synchronized void save() throws java.io.IOException {
    //TRACE (".save().");

    if(!changed) return;

    if(urlstring==null) return;


    if(urlstring.startsWith("file://")) {
      try {
        java.io.File f = new java.io.File(urlstring.substring(7));
        f.delete();
        OutputStream output=new java.io.FileOutputStream(urlstring.substring(7));

        JAXBContext jc = JAXBContext.newInstance("fips.sim.dig.edit.rcfile");
        Marshaller m = jc.createMarshaller();
        m.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
        m.marshal(content, output);
        output.close();
        changed=false;
      } catch (JAXBException jbe) {
        jbe.printStackTrace();
        throw new IOException("Marshalling of xml-data failed");
      }
    }
  }




  public void run() {
    while(true) {
      if(autowrite&&changed) {
        try {
          save();
        } catch (java.io.IOException ioe) {
          ioe.printStackTrace();
          changed=false;
        }
      } else {
        synchronized(this) {
          try {
            wait(3000);
          } catch (InterruptedException ie) {
          }
        }
      }
    }
  }



  protected synchronized void finalize() {
    //TRACE (".finalize().");

    if(autowrite) {
      try {
        save();
      } catch (java.io.IOException ioe) {
        ioe.printStackTrace();
      }
    }
  }



}












